"""Utils for logging."""

import io
import logging
import os
import sys

from .. import _logger


formatter_debug = logging.Formatter("%(asctime)s\t[%(levelname)s]\t%(pathname)s.%(funcName)s:%(lineno)d\t%(message)s")
formatter_default = logging.Formatter("[%(asctime)s] %(message)s", "%H:%M:%S")

INFO2 = 17
INFO3 = 13


def add_logging_level(levelName, levelNum, methodName=None):
    """Comprehensively adds a new logging level to the `logging` module and the currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example:
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5


    Args:
        levelName: Level name.
        levelNum: Level number.
        methodName: Method name.

    """
    assert (levelNum > 0) and (levelNum < 50)
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError("{} already defined in logging module".format(levelName))
    if hasattr(logging, methodName):
        raise AttributeError("{} already defined in logging module".format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError("{} already defined in logger class".format(methodName))

    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


add_logging_level("INFO2", INFO2)
add_logging_level("INFO3", INFO3)


class LoggerStream(io.IOBase):
    def __init__(self, logger, verbose_eval=100) -> None:
        super().__init__()
        self.logger = logger
        self.verbose_eval = verbose_eval
        self.counter = 1

    def write(self, message):
        if message == "\n":
            return
        iter_num = message.split("\t")[0]
        if (iter_num == "[1]") or (iter_num == "0:") or ((iter_num[-1] != "]") and (iter_num[-1] != ":")):
            self.logger.info3(message.rstrip())
            return

        if self.counter < self.verbose_eval - 1:
            self.logger.debug(message.rstrip())
            self.counter += 1
        else:
            self.logger.info3(message.rstrip())
            self.counter = 0


def verbosity_to_loglevel(verbosity: int, extended=True):
    if extended:
        if verbosity <= 0:
            log_level = logging.ERROR
        elif verbosity == 1:
            log_level = logging.INFO
        elif verbosity == 2:
            log_level = logging.INFO2
        elif verbosity == 3:
            log_level = logging.INFO3
        else:
            log_level = logging.DEBUG
    else:
        if verbosity <= 0:
            log_level = logging.ERROR
        elif verbosity == 1:
            log_level = logging.INFO
        else:
            log_level = logging.DEBUG

    return log_level


def get_stdout_level():
    for handler in _logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            return handler.level
    return _logger.getEffectiveLevel()


def set_stdout_level(level):
    _logger.setLevel(logging.DEBUG)

    has_console_handler = False

    for handler in _logger.handlers:
        if isinstance(handler, logging.StreamHandler):
            if handler.level == level:
                has_console_handler = True
            else:
                _logger.handlers.remove(handler)

    if not has_console_handler:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter_default)
        handler.setLevel(level)

        _logger.addHandler(handler)


def add_filehandler(filename: str, level=logging.DEBUG):
    if filename:
        has_file_handler = False

        for handler in _logger.handlers:
            if isinstance(handler, logging.FileHandler):
                if handler.baseFilename == filename or handler.baseFilename == os.path.join(os.getcwd(), filename):
                    has_file_handler = True
                else:
                    _logger.handlers.remove(handler)

        if not has_file_handler:
            file_handler = logging.FileHandler(filename, mode="w")

            if level == logging.DEBUG:
                file_handler.setFormatter(formatter_debug)
            else:
                file_handler.setFormatter(formatter_default)

            file_handler.setLevel(level)

            # if handler_filter:
            #     file_handler.addFilter(handler_filter)

            _logger.addHandler(file_handler)
    else:
        for handler in _logger.handlers:
            if isinstance(handler, logging.FileHandler):
                _logger.handlers.remove(handler)


class DuplicateFilter(object):
    def __init__(self):
        self.msgs = set()

    def filter(self, record):
        rv = record.msg not in self.msgs
        self.msgs.add(record.msg)
        return rv



# Standard python libraries
import logging
import os

import yaml


logging.basicConfig(format="[%(asctime)s] (%(levelname)s): %(message)s", level=logging.INFO)

# Installed libraries
import numpy as np
import pandas as pd

from statsmodels.tsa.seasonal import STL
from statsmodels.tsa.seasonal import seasonal_decompose

# Imports from our package
from ...automl.base import AutoML
from ...automl.presets.base import upd_params
from ...automl.presets.tabular_presets import TabularAutoML
from ...dataset.roles import DatetimeRole
from ...ml_algo.linear_sklearn import LinearLBFGS
from ...pipelines.features.linear_pipeline import LinearTrendFeatures
from ...pipelines.ml.base import MLPipeline
from ...reader.base import DictToPandasSeqReader
from ...tasks import Task


class TrendModel:

    _available_trend_types = ["decompose", "decompose_STL", "linear", "rolling"]

    def __init__(self, params=None):
        self.params = params
        assert self.params["trend_type"] in self._available_trend_types

    def _detect_step(self, x):
        x_min, x_max = tuple(
            np.quantile(x, [self.params["detect_step_quantile"], 1 - self.params["detect_step_quantile"]])
        )
        x_range = x_max - x_min
        window = self.params["detect_step_window"]
        diff = np.zeros(len(x) - 2 * window)
        for i in range(len(x) - 2 * window):
            diff[i] = np.median(x[i + window : i + 2 * window]) - np.median(x[i : i + window])
        diff = np.abs(diff) / x_range
        diff = np.concatenate((diff[0] * np.ones(window), diff, diff[-1] * np.ones(window)))
        return np.any(diff > self.params["detect_step_threshold"])

    def _get_rolling_median(self, train_data, roles):
        median = train_data[roles["target"]].rolling(self.params["rolling_size"]).apply(np.median)
        return median.fillna(median[~median.isna()].values[0]).values

    def _detect_no_trend(self, x):
        pass

    def _estimate_trend(self, train_data, roles):
        if self.params["trend_type"] == "decompose":
            return seasonal_decompose(
                train_data[roles["target"]].values,
                model="additive",
                period=self.params["decompose_period"],
                extrapolate_trend="freq",
            ).trend
        elif self.params["trend_type"] == "decompose_STL":
            return STL(train_data[roles["target"]].values, period=self.params["decompose_period"]).fit().trend
        elif self.params["trend_type"] == "rolling":
            return self._get_rolling_median(train_data, roles)

    def fit_predict(self, train_data, roles):
        """Fit and predict data.

        if self._detect_no_trend(train_data[roles['target']]):
            self.params['trend'] = False

        # noqa: DAR101
        # noqa: DAR201

        """
        self.roles = roles
        if not self.params["trend"]:
            return np.zeros(len(train_data))
        if self._detect_step(train_data[roles["target"]]):
            self.params["trend_type"] = "rolling"

        task_trend = Task("reg", greater_is_better=False, metric="mae", loss="mae")
        reader_trend = DictToPandasSeqReader(task=task_trend, cv=2, seq_params={})
        reader_trend.fit_read({"plain": train_data, "seq": None}, roles=roles)
        timerole = [key for key, value in reader_trend._roles.items() if isinstance(value, DatetimeRole)][0]

        feats_trend = LinearTrendFeatures()
        model_trend = LinearLBFGS()
        pipeline_trend = MLPipeline(
            [model_trend], pre_selection=None, features_pipeline=feats_trend, post_selection=None
        )
        self.automl_trend = AutoML(reader_trend, [[pipeline_trend]], skip_conn=False)

        if self.params["trend_type"] in ["decompose", "decompose_STL", "rolling"]:
            trend = self._estimate_trend(train_data, roles)
            if self.params["train_on_trend"]:
                trend_data = train_data[[roles["target"], timerole]].copy()
                trend_data.drop(roles["target"], axis=1, inplace=True)
                trend_data["trend"] = trend
                roles = {"target": "trend"}
                _ = self.automl_trend.fit_predict(
                    {"plain": trend_data.iloc[-self.params["trend_size"] :], "seq": None}, roles=roles
                )
            else:
                _ = self.automl_trend.fit_predict(
                    {"plain": train_data[[roles["target"], timerole]].iloc[-self.params["trend_size"] :], "seq": None},
                    roles=roles,
                )

        elif self.params["trend_type"] == "linear":
            _ = self.automl_trend.fit_predict(
                {"plain": train_data[[roles["target"], timerole]], "seq": None}, roles=roles
            )
            trend = self.automl_trend.predict({"plain": train_data, "seq": None}).data[:, 0]
        return trend

    def predict(self, data, future_time):
        MIN_PREDICT_HISTORY = 5 * self.params["trend_size"]
        if not self.params["trend"]:
            return np.zeros(len(data)), np.zeros(len(future_time))
        if self.params["trend_type"] == "linear" or len(data) < MIN_PREDICT_HISTORY:
            trend = self.automl_trend.predict({"plain": data, "seq": None}).data[:, 0]
        else:
            trend = self._estimate_trend(data, self.roles)
        pred_trend = self.automl_trend.predict({"plain": future_time, "seq": None}).data[:, 0]
        return trend, pred_trend


class AutoTS:
    @property
    def n_target(self):
        """Get length of future prediction.

        Returns:
            length
        """
        return self.reader_params["seq_params"]["seq0"]["params"]["n_target"]

    @property
    def n_history(self):
        """Get length of history used for feature generation.

        Returns:
            length
        """
        return self.reader_params["seq_params"]["params"]["history"]

    @property
    def datetime_key(self):
        """Get name of datetime index column

        Returns:
            column name
        """
        return (
            self.TM.automl_trend.levels[0][0]
            .features_pipeline._pipeline.transformer_list[0]
            .transformer_list[0]
            .keys[0]
        )

    def __init__(self, task, time_series_trend_params=None, reader_params=None, **kwargs):
        self.task = task
        self.task_trend = Task("reg", greater_is_better=False, metric="mae", loss="mae")
        self.kwargs = kwargs

        if "config_path" not in kwargs:
            _base_dir = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]
            self.kwargs["config_path"] = os.path.join(_base_dir, "automl", "presets", "time_series_config.yml")

        with open(self.kwargs["config_path"]) as f:
            params = yaml.safe_load(f)

        # TrendModel and reader_params initialization
        for name, param in zip(
            ["time_series_trend_params", "reader_params"], [time_series_trend_params, reader_params]
        ):
            if param is None:
                param = {}
            self.__dict__[name] = upd_params(params[name], param)

        self.TM = TrendModel(params=self.time_series_trend_params)

    def fit_predict(self, train_data, roles, verbose=0):
        self.roles = roles
        train_trend = self.TM.fit_predict(train_data, roles)

        if hasattr(self.TM, "automl_trend"):
            self.datetime_step = (
                pd.to_datetime(train_data[self.datetime_key]).iloc[1]
                - pd.to_datetime(train_data[self.datetime_key]).iloc[0]
            )
        # fit main
        train_detrend = train_data.copy()
        train_detrend.loc[:, roles["target"]] = train_detrend.loc[:, roles["target"]] - train_trend

        # Tabular preset
        self.automl_seq = TabularAutoML(
            task=self.task, is_time_series=True, reader_params=self.reader_params, **self.kwargs
        )

        oof_pred_seq = self.automl_seq.fit_predict({"seq": {"seq0": train_detrend}}, roles=roles, verbose=verbose)
        return oof_pred_seq, train_trend

    def predict(self, data, return_raw=False):
        test_idx = None
        if self.time_series_trend_params["trend"] is True:
            last_datetime = pd.to_datetime(data[self.datetime_key]).values[-1]
            vals = [last_datetime + (i + 1) * self.datetime_step for i in range(self.n_target)]
            if not self.reader_params["seq_params"]["seq0"]["params"]["test_last"]:
                vals = data[self.datetime_key].tolist() + vals
            test_data = pd.DataFrame(vals, columns=[self.datetime_key])
            if not self.reader_params["seq_params"]["seq0"]["params"]["test_last"]:
                test_idx = self.automl_seq.reader.ti["seq0"].create_target(test_data, plain_data=None)
            trend, test_pred_trend = self.TM.predict(data, test_data)
        else:
            test_pred_trend = np.zeros(self.n_target)
            trend = np.zeros(len(data))

        detrend = data.copy()
        detrend.loc[:, self.roles["target"]] = detrend.loc[:, self.roles["target"]] - trend
        test_pred_detrend = self.automl_seq.predict({"seq": {"seq0": detrend}})
        if return_raw:
            return test_pred_detrend

        if test_pred_detrend.data.shape[0] == 1:
            final_pred = test_pred_trend + test_pred_detrend.data.flatten()
        else:
            if (test_idx is not None) and (not self.reader_params["seq_params"]["seq0"]["params"]["test_last"]):
                test_pred_trend = test_pred_trend[test_idx]
            final_pred = test_pred_trend + test_pred_detrend.data
        return final_pred, test_pred_trend



"""Metrics and loss functions for XGBoost."""

import logging

from functools import partial
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import xgboost

from ..common_metric import _valid_str_multiclass_metric_names
from ..utils import infer_gib
from .base import Loss
from .base import fw_rmsle
from .lgb_custom import lgb_f1_loss_multiclass  # , F1Factory
from .lgb_custom import softmax_ax1


logger = logging.getLogger(__name__)


_xgb_binary_metrics_dict = {
    "auc": "auc",
    "logloss": "logloss",
    "accuracy": "error",
}

_xgb_reg_metrics_dict = {
    "mse": "rmse",
    "mae": "mae",
    "r2": "rmse",
    "rmsle": "rmsle",
    "mape": "mape",
    # "quantile": "", # TODO
    "huber": "mphe",
    # "fair": "",# TODO
}

_xgb_multiclass_metrics_dict = {
    "auc": _valid_str_multiclass_metric_names["auc"],
    "auc_mu": _valid_str_multiclass_metric_names["auc_mu"],
    "crossentropy": "mlogloss",
    "accuracy": "merror",
    "f1_macro": _valid_str_multiclass_metric_names["f1_macro"],
    "f1_micro": _valid_str_multiclass_metric_names["f1_micro"],
    "f1_weighted": _valid_str_multiclass_metric_names["f1_weighted"],
}

_xgb_metrics_dict = {
    "binary": _xgb_binary_metrics_dict,
    "reg": _xgb_reg_metrics_dict,
    "multiclass": _xgb_multiclass_metrics_dict,
}

_xgb_loss_mapping = {
    "logloss": ("binary:logistic", None, None),
    "mse": ("reg:squarederror", None, None),
    "mae": ("reg:absoluteerror", None, None),
    # "mape": ("mape", None, None),
    "crossentropy": ("multi:softmax", None, None),  # TODO
    "rmsle": ("reg:squaredlogerror", fw_rmsle, np.expm1),  # TODO: fw and bw
    "quantile": ("reg:quantileerror", None, None),
    "huber": ("reg:pseudohubererror", None, None),
    # "fair": ("fair", None, None),
    "f1": (lgb_f1_loss_multiclass, None, softmax_ax1),
}

_xgb_loss_params_mapping = {
    "quantile": {"q": "quantile_alpha"},
    "huber": {"a": "huber_slope"},
    # "fair_c": {"c": "fair_c"},# TODO
}

_xgb_force_metric = {
    "rmsle": ("mse", None, None),  # TODO
}


class XGBFunc:
    """Wrapper of metric function for LightGBM."""

    def __init__(self, metric_func, greater_is_better, bw_func):
        self.metric_func = metric_func
        self.greater_is_better = greater_is_better
        self.bw_func = bw_func

    def __call__(self, pred: np.ndarray, dtrain: xgboost.DMatrix) -> Tuple[str, float, bool]:
        """Calculate metric."""
        label = dtrain.get_label()

        weights = dtrain.get_weight()

        if label.shape[0] != pred.shape[0]:
            pred = pred.reshape((label.shape[0], -1), order="F")
            label = label.astype(np.int32)

        label = self.bw_func(label)
        pred = self.bw_func(pred)

        # for weighted case
        try:
            val = self.metric_func(label, pred, sample_weight=weights)
        except TypeError:
            val = self.metric_func(label, pred)

        # TODO: what if grouped case

        return "Opt metric", val, self.greater_is_better


class XGBLoss(Loss):
    """Loss used for XGBoost.

    Args:
        loss: Objective to optimize.
        loss_params: additional loss parameters.
            Format like in :mod:`lightautoml.tasks.custom_metrics`.
        fw_func: forward transformation.
            Used for transformation of target and item weights.
        bw_func: backward transformation.
            Used for predict values transformation.

    Note:
        Loss can be one of the types:

            - Str: one of default losses
                ('auc', 'mse', 'mae', 'logloss', 'accuray', 'r2',
                'rmsle', 'mape', 'quantile', 'huber', 'fair')
                or another XGBoost objective.
            - Callable: custom XGBoost style objective.

    """

    def __init__(
        self,
        loss: Union[str, Callable],
        loss_params: Optional[Dict] = None,
        fw_func: Optional[Callable] = None,
        bw_func: Optional[Callable] = None,
    ):
        if loss in _xgb_loss_mapping:
            fobj, fw_func, bw_func = _xgb_loss_mapping[loss]
            if type(fobj) is str:
                self.fobj_name = fobj
                self.fobj = None
            else:
                self.fobj_name = None
                self.fobj = fobj
            # map param name for known objectives
            if self.fobj_name in _xgb_loss_params_mapping:
                param_mapping = _xgb_loss_params_mapping[self.fobj_name]
                loss_params = {param_mapping[x]: loss_params[x] for x in loss_params}

        else:
            # set lgb style objective
            if type(loss) is str:
                self.fobj_name = loss
                self.fobj = None
            else:
                self.fobj_name = None
                self.fobj = loss

        # set forward and backward transformations
        if fw_func is not None:
            self._fw_func = fw_func
        if bw_func is not None:
            self._bw_func = bw_func

        self.fobj_params = {}
        if loss_params is not None:
            self.fobj_params = loss_params

        self.metric = None

    def metric_wrapper(
        self,
        metric_func: Callable,
        greater_is_better: Optional[bool],
        metric_params: Optional[Dict] = None,
    ) -> Callable:
        """Customize metric.

        Args:
            metric_func: Callable metric.
            greater_is_better: Whether or not higher value is better.
            metric_params: Additional metric parameters.

        Returns:
            Callable metric, that returns ('Opt metric', value, greater_is_better).

        """
        if greater_is_better is None:
            greater_is_better = infer_gib(metric_func)

        if metric_params is not None:
            metric_func = partial(metric_func, **metric_params)

        return XGBFunc(metric_func, greater_is_better, self._bw_func)

    def set_callback_metric(
        self,
        metric: Union[str, Callable],
        greater_is_better: Optional[bool] = None,
        metric_params: Optional[Dict] = None,
        task_name: Optional[str] = None,
    ):
        """Callback metric setter.

        Args:
            metric: Callback metric.
            greater_is_better: Whether or not higher value is better.
            metric_params: Additional metric parameters.
            task_name: Name of task.

        Note:
            Value of ``task_name`` should be one of following options:

            - `'binary'`
            - `'reg'`
            - `'multiclass'`

        """
        # force metric if special loss
        # what about task_name? in this case?
        if self.fobj_name in _xgb_force_metric:
            metric, greater_is_better, metric_params = _xgb_force_metric[self.fobj_name]
            logger.info2(
                "For xgb {0} callback metric switched to {1}".format(self.fobj_name, metric),
                UserWarning,
            )

        self.metric_params = {}

        # set lgb style metric
        self.metric = metric
        if type(metric) is str:

            if metric_params is not None:
                self.metric_params = metric_params

            _metric_dict = _xgb_metrics_dict[task_name]
            _metric = _metric_dict.get(metric)
            if type(_metric) is str:
                self.metric_name = _metric
                self.feval = None
            else:
                self.metric_name = None
                # _metric = CustomWrapper(_metric)
                self.feval = self.metric_wrapper(_metric, greater_is_better, {})

        else:
            self.metric_name = None
            # metric = CustomWrapper(metric)
            self.feval = self.metric_wrapper(metric, greater_is_better, self.metric_params)



"""."""

from typing import Callable

import numpy as np


def infer_gib(metric: Callable) -> bool:
    """Infer greater is better from metric.

    Args:
        metric: Score or loss function.

    Returns:
        ```True``` if grater is better.

    Raises:
        AssertionError: If there is no way to order the predictions.

    """
    label = np.array([0, 1])
    pred = np.array([0.1, 0.9])

    g_val = metric(label, pred)
    b_val = metric(label, pred[::-1])

    assert g_val != b_val, "Cannot infer greater is better from metric." " Should be set manually."

    return g_val > b_val


def infer_gib_multiclass(metric: Callable) -> bool:
    """Infer greater is better from metric.

    Args:
        metric: Metric function. It must take two arguments y_true, y_pred.

    Returns:
        ```True``` if grater is better.

    Raises:
        AssertionError: If there is no way to order the predictions.

    """
    label = np.array([0, 1, 2])
    pred = np.array([[0.9, 0.05, 0.05], [0.05, 0.9, 0.05], [0.05, 0.05, 0.9]])

    g_val = metric(label, pred)
    b_val = metric(label, pred[::-1])

    assert g_val != b_val, "Cannot infer greater is better from metric. " "Should be set manually."

    return g_val > b_val



"""Tools for partial installation."""

import os


try:
    from importlib.metadata import PackageNotFoundError
    from importlib.metadata import distribution
except ModuleNotFoundError:
    from importlib_metadata import PackageNotFoundError, distribution

import logging


logger = logging.getLogger(__name__)


def __validate_extra_deps(extra_section: str, error: bool = False) -> None:
    """Check if extra dependencies is installed.

    Args:
        extra_section: Name of extra dependencies
        error: How to process error

    """
    ignore_deps = os.environ.get("DOCUMENTATION_ENV", False)

    md = distribution("lightautoml").metadata
    extra_pattern = 'extra == "{}"'.format(extra_section)
    reqs_info = []
    for k, v in md.items():
        if k == "Requires-Dist" and extra_pattern in v:
            req = v.split(";")[0].split()[0]
            reqs_info.append(req)

    for req_info in reqs_info:
        lib_name: str = req_info.split()[0]
        try:
            distribution(lib_name)
        except PackageNotFoundError as e:
            # Print warning
            logger.warning(
                "'%s' extra dependency package '%s' isn't installed. "
                "Look at README.md in repo 'LightAutoML' for installation instructions.",
                extra_section,
                lib_name,
            )

            if not ignore_deps:
                if error:
                    raise e



"""AutoML presets for data with texts."""

from lightautoml.utils.installation import __validate_extra_deps


__validate_extra_deps("nlp", error=True)


import logging
import os

from copy import deepcopy
from typing import Optional
from typing import Sequence

import torch
import torch.nn as nn

from pandas import DataFrame

from ...ml_algo.boost_cb import BoostCB
from ...ml_algo.boost_lgbm import BoostLGBM
from ...ml_algo.dl_model import TorchModel
from ...ml_algo.linear_sklearn import LinearLBFGS
from ...ml_algo.tuning.optuna import OptunaTuner
from ...pipelines.features.base import FeaturesPipeline
from ...pipelines.features.lgb_pipeline import LGBAdvancedPipeline
from ...pipelines.features.linear_pipeline import LinearFeatures
from ...pipelines.features.text_pipeline import NLPTFiDFFeatures
from ...pipelines.features.text_pipeline import TextAutoFeatures
from ...pipelines.features.text_pipeline import TextBertFeatures
from ...pipelines.ml.nested_ml_pipe import NestedTabularMLPipeline
from ...pipelines.selection.base import SelectionPipeline
from ...reader.base import PandasToPandasReader
from ...reader.tabular_batch_generator import ReadableToDf
from ...tasks import Task
from ..blend import WeightedBlender
from .base import upd_params
from .tabular_presets import NumpyDataset
from .tabular_presets import TabularAutoML


logger = logging.getLogger(__name__)


_base_dir = os.path.dirname(__file__)
# set initial runtime rate guess for first level models
_time_scores = {
    "lgb": 1,
    "lgb_tuned": 3,
    "linear_l2": 0.7,
    "cb": 2,
    "cb_tuned": 6,
    "nn": 10,
    "nn_tuned": 20,
    "rf": 5,
    "rf_tuned": 10,
}


# TODO: add text feature selection
class TabularNLPAutoML(TabularAutoML):
    """Classic preset - work with tabular and text data.

    Supported data roles - numbers, dates, categories, text
    Limitations - no memory management.

    GPU support in catboost/lightgbm (if installed for GPU),
    NN models training.

    Commonly _params kwargs (ex. timing_params) set via
    config file (config_path argument).
    If you need to change just few params, it's possible to pass
    it as dict of dicts, like json.
    To get available params please look on default config template.
    Also you can find there param description.
    To generate config template call
    :meth:`TabularNLPAutoML.get_config('config_path.yml')`.


    Args:
        task: Task to solve.
        timeout: Timeout in seconds.
        memory_limit: Memory limit that are passed to each automl.
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.
        debug: To catch running model exceptions or not.
        timing_params: Timing param dict.
        config_path: Path to config file.
        general_params: General param dict.
        reader_params: Reader param dict.
        read_csv_params: Params to pass :func:`pandas.read_csv`
            (case of train/predict from file).
        nested_cv_params: Param dict for nested cross-validation.
        tuning_params: Params of Optuna tuner.
        selection_params: Params of feature selection.
        nn_params: Params of neural network model.
        lgb_params: Params of lightgbm model.
        cb_params: Params of catboost model.
        linear_l2_params: Params of linear model.
        nn_pipeline_params: Params of feature generation
            for neural network models.
        gbm_pipeline_params: Params of feature generation
            for boosting models.
        linear_pipeline_params: Params of feature generation
            for linear models.
        text_params: General params of text features.
        tfidf_params: Params of tfidf features.
        autonlp_params: Params of text embeddings features.

    """

    _default_config_path = "text_config.yml"

    _time_scores = {
        "lgb": 1,
        "lgb_tuned": 3,
        "linear_l2": 0.7,
        "cb": 2,
        "cb_tuned": 6,
        "nn": 1,
        "rf": 5,
        "rf_tuned": 10,
    }

    def __init__(
        self,
        task: Task,
        timeout: int = 3600,
        memory_limit: int = 16,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
        debug: bool = False,
        timing_params: Optional[dict] = None,
        config_path: Optional[str] = None,
        general_params: Optional[dict] = None,
        reader_params: Optional[dict] = None,
        read_csv_params: Optional[dict] = None,
        nested_cv_params: Optional[dict] = None,
        tuning_params: Optional[dict] = None,
        selection_params: Optional[dict] = None,
        nn_params: Optional[dict] = None,
        lgb_params: Optional[dict] = None,
        cb_params: Optional[dict] = None,
        rf_params: Optional[dict] = None,
        linear_l2_params: Optional[dict] = None,
        nn_pipeline_params: Optional[dict] = None,
        gbm_pipeline_params: Optional[dict] = None,
        linear_pipeline_params: Optional[dict] = None,
        text_params: Optional[dict] = None,
        tfidf_params: Optional[dict] = None,
        autonlp_params: Optional[dict] = None,
    ):
        super().__init__(
            task,
            timeout,
            memory_limit,
            cpu_limit,
            gpu_ids,
            debug,
            timing_params,
            config_path,
        )

        # upd manual params
        for name, param in zip(
            [
                "general_params",
                "reader_params",
                "read_csv_params",
                "nested_cv_params",
                "tuning_params",
                "selection_params",
                "nn_params",
                "lgb_params",
                "cb_params",
                "rf_params",
                "linear_l2_params",
                "nn_pipeline_params",
                "gbm_pipeline_params",
                "linear_pipeline_params",
                "text_params",
                "tfidf_params",
                "autonlp_params",
            ],
            [
                general_params,
                reader_params,
                read_csv_params,
                nested_cv_params,
                tuning_params,
                selection_params,
                nn_params,
                lgb_params,
                cb_params,
                rf_params,
                linear_l2_params,
                nn_pipeline_params,
                gbm_pipeline_params,
                linear_pipeline_params,
                text_params,
                tfidf_params,
                autonlp_params,
            ],
        ):
            if param is None:
                param = {}
            self.__dict__[name] = upd_params(self.__dict__[name], param)

    def infer_auto_params(self, train_data: DataFrame, multilevel_avail: bool = False):

        # infer gpu params
        gpu_cnt = torch.cuda.device_count()
        gpu_ids = self.gpu_ids
        if gpu_cnt > 0 and gpu_ids:
            if gpu_ids == "all":
                gpu_ids = ",".join(list(map(str, range(gpu_cnt))))

            self.nn_params["device"] = gpu_ids.split(",")
            self.text_params["device"] = gpu_ids.split(",")

            if self.general_params["use_algos"] == "auto":
                self.general_params["use_algos"] = [["linear_l2", "cb", "nn"]]

        else:
            self.nn_params["device"] = "cpu"
            self.text_params["device"] = "cpu"

            if self.general_params["use_algos"] == "auto":
                self.general_params["use_algos"] = [["linear_l2", "lgb"]]

        # check all n_jobs params
        cpu_cnt = min(os.cpu_count(), self.cpu_limit)
        torch.set_num_threads(cpu_cnt)

        self.nn_params["num_workers"] = min(self.nn_params["num_workers"], cpu_cnt)
        self.nn_params["lang"] = self.nn_params["lang"] or self.text_params["lang"]
        self.nn_params["bert_name"] = self.nn_params["bert_name"] or self.text_params["bert_model"]

        logger.info3("Model language mode: {}".format(self.nn_params["lang"]))

        if isinstance(self.autonlp_params["transformer_params"], dict):
            if "loader_params" in self.autonlp_params["transformer_params"]:
                self.autonlp_params["transformer_params"]["loader_params"]["num_workers"] = min(
                    self.autonlp_params["transformer_params"]["loader_params"]["num_workers"],
                    cpu_cnt,
                )
            else:
                self.autonlp_params["transformer_params"]["loader_params"] = {"num_workers": cpu_cnt}

        # other params as tabular
        super().infer_auto_params(train_data, multilevel_avail)

    def get_nlp_pipe(self, type: str = "tfidf") -> Optional[FeaturesPipeline]:
        if type == "tfidf":
            return NLPTFiDFFeatures(**self.text_params, **self.tfidf_params)
        elif type == "embed":
            return TextAutoFeatures(**self.text_params, **self.autonlp_params)
        elif type == "bert":
            return TextBertFeatures(**self.text_params)
        else:
            return None

    def get_nn(
        self, keys: Sequence[str], n_level: int = 1, pre_selector: Optional[SelectionPipeline] = None
    ) -> NestedTabularMLPipeline:
        ml_algos = []
        force_calc = []

        text_nn_feats = self.get_nlp_pipe(self.nn_pipeline_params["text_features"])
        nn_feats = LinearFeatures(output_categories=True, **self.linear_pipeline_params)
        if text_nn_feats is not None:
            nn_feats.append(text_nn_feats)

        general_nn_params = deepcopy(self.nn_params)
        if "0" in self.nn_params:
            for i in range(len(keys)):
                if str(i) in general_nn_params:
                    del general_nn_params[str(i)]

        for i, key in enumerate(keys):
            time_score = self.get_time_score(n_level, "nn")
            nn_timer = self.timer.get_task_timer("reg_nn", time_score)
            model_params = deepcopy(general_nn_params)

            model_params.update(self.nn_params.get(str(i), general_nn_params))
            model_name = key

            if isinstance(key, str):
                tuned = "_tuned" in key
                if key[:2] == "nn":
                    model_name = "_linear_layer"
                _name = "TorchNN_" + model_name + "_" + str(i)
                model_name = model_name.replace("_tuned", "")
            else:
                tuned = model_params.get("tuned")
                _name = "TorchNN_" + str(i)

            model_params["model"] = model_name
            nn_model = TorchModel(
                timer=nn_timer,
                default_params=model_params,
                optimization_search_space=model_params.get("optimization_search_space", None),
            )
            nn_model._name = _name

            if tuned:
                nn_model.set_prefix("Tuned")
                nn_tuner = OptunaTuner(
                    n_trials=model_params["tuning_params"]["max_tuning_iter"],
                    timeout=model_params["tuning_params"]["max_tuning_time"],
                    fit_on_holdout=model_params["tuning_params"]["fit_on_holdout"],
                )
                nn_model = (nn_model, nn_tuner)
            ml_algos.append(nn_model)
            force_calc.append(True if not len(ml_algos) - 1 else False)

        nn_pipe = NestedTabularMLPipeline(
            ml_algos, force_calc, pre_selection=None, features_pipeline=nn_feats, **self.nested_cv_params
        )

        return nn_pipe

    def get_linear(self, n_level: int = 1, pre_selector: Optional[SelectionPipeline] = None) -> NestedTabularMLPipeline:

        # linear model with l2
        time_score = self.get_time_score(n_level, "linear_l2")
        linear_l2_timer = self.timer.get_task_timer("reg_l2", time_score)
        linear_l2_model = LinearLBFGS(timer=linear_l2_timer, **self.linear_l2_params)

        text_l2_feats = self.get_nlp_pipe(self.linear_pipeline_params["text_features"])
        linear_l2_feats = LinearFeatures(output_categories=True, **self.linear_pipeline_params)
        if text_l2_feats is not None:
            linear_l2_feats.append(text_l2_feats)

        linear_l2_pipe = NestedTabularMLPipeline(
            [linear_l2_model],
            force_calc=True,
            pre_selection=pre_selector,
            features_pipeline=linear_l2_feats,
            **self.nested_cv_params
        )
        return linear_l2_pipe

    def get_gbms(
        self,
        keys: Sequence[str],
        n_level: int = 1,
        pre_selector: Optional[SelectionPipeline] = None,
    ):

        text_gbm_feats = self.get_nlp_pipe(self.gbm_pipeline_params["text_features"])
        gbm_feats = LGBAdvancedPipeline(feats_imp=pre_selector, output_categories=False, **self.gbm_pipeline_params)
        if text_gbm_feats is not None:
            gbm_feats.append(text_gbm_feats)

        ml_algos = []
        force_calc = []
        for key, force in zip(keys, [True, False, False, False]):
            tuned = "_tuned" in key
            algo_key = key.split("_")[0]
            time_score = self.get_time_score(n_level, key)
            gbm_timer = self.timer.get_task_timer(algo_key, time_score)
            if algo_key == "lgb":
                gbm_model = BoostLGBM(timer=gbm_timer, **self.lgb_params)
            elif algo_key == "cb":
                gbm_model = BoostCB(timer=gbm_timer, **self.cb_params)
            else:
                raise ValueError("Wrong algo key")

            if tuned:
                gbm_tuner = OptunaTuner(
                    n_trials=self.tuning_params["max_tuning_iter"],
                    timeout=self.tuning_params["max_tuning_time"],
                    fit_on_holdout=self.tuning_params["fit_on_holdout"],
                )
                gbm_model = (gbm_model, gbm_tuner)
            ml_algos.append(gbm_model)
            force_calc.append(force)

        gbm_pipe = NestedTabularMLPipeline(
            ml_algos, force_calc, pre_selection=pre_selector, features_pipeline=gbm_feats, **self.nested_cv_params
        )

        return gbm_pipe

    def create_automl(self, **fit_args):
        """Create basic automl instance.

        Args:
            **fit_args: Contain all information needed for creating automl.

        """
        train_data = fit_args["train_data"]
        self.infer_auto_params(train_data)
        reader = PandasToPandasReader(task=self.task, **self.reader_params)

        pre_selector = self.get_selector()

        levels = []

        for n, names in enumerate(self.general_params["use_algos"]):
            lvl = []
            # regs
            if "linear_l2" in names:
                selector = None
                if "linear_l2" in self.selection_params["select_algos"] and (
                    self.general_params["skip_conn"] or n == 0
                ):
                    selector = pre_selector
                lvl.append(self.get_linear(n + 1, selector))

            gbm_models = [
                x for x in ["lgb", "lgb_tuned", "cb", "cb_tuned"] if x in names and x.split("_")[0] in self.task.losses
            ]

            if len(gbm_models) > 0:
                selector = None
                if "gbm" in self.selection_params["select_algos"] and (self.general_params["skip_conn"] or n == 0):
                    selector = pre_selector
                lvl.append(self.get_gbms(gbm_models, n + 1, selector))

            available_nn_models = ["nn", "mlp", "dense", "denselight", "resnet", "snn", "linear_layer", "_linear_layer"]
            available_nn_models = available_nn_models + [x + "_tuned" for x in available_nn_models]
            nn_models = [
                x for x in names if x in available_nn_models or (isinstance(x, type) and issubclass(x, nn.Module))
            ]

            if len(nn_models) > 0:
                selector = None
                lvl.append(self.get_nn(nn_models, n + 1, selector))

            levels.append(lvl)

        # blend everything
        blender = WeightedBlender()

        # initialize
        self._initialize(
            reader,
            levels,
            skip_conn=self.general_params["skip_conn"],
            blender=blender,
            timer=self.timer,
            debug=self.debug,
        )

    def predict(
        self,
        data: ReadableToDf,
        features_names: Optional[Sequence[str]] = None,
        batch_size: Optional[int] = None,
        n_jobs: Optional[int] = 1,
    ) -> NumpyDataset:
        """Get dataset with predictions.

        Almost same as :meth:`lightautoml.automl.base.AutoML.predict`
        on new dataset, with additional features.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`. For example,
              ``{'data': X...}``. In this case roles are optional,
              but `train_features` and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Parallel inference - you can pass ``n_jobs`` to speedup
        prediction (requires more RAM).
        Batch_inference - you can pass ``batch_size``
        to decrease RAM usage (may be longer).

        Args:
            data: Dataset to perform inference.
            features_names: Optional features names,
                if cannot be inferred from `train_data`.
            batch_size: Batch size or ``None``.
            n_jobs: Number of jobs.

        Returns:
            Dataset with predictions.

        """
        return super().predict(data, features_names, batch_size)



"""Blenders."""

import logging

from typing import Callable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import cast
from typing import Dict

import numpy as np

from scipy.optimize import minimize_scalar

from ..dataset.base import LAMLDataset
from ..dataset.np_pd_dataset import NumpyDataset
from ..dataset.roles import NumericRole
from ..pipelines.ml.base import MLPipeline


logger = logging.getLogger(__name__)

np.seterr(divide="ignore", invalid="ignore")


class Blender:
    """Basic class for blending.

    Blender learns how to make blend
    on sequence of prediction datasets and prune pipes,
    that are not used in final blend.

    """

    _outp_dim = None

    _bypass = False

    @property
    def outp_dim(self) -> int:  # noqa: D102
        return self._outp_dim

    def fit_predict(
        self, predictions: Sequence[LAMLDataset], pipes: Sequence[MLPipeline], class_mapping: Dict
    ) -> Tuple[LAMLDataset, Sequence[MLPipeline]]:
        """Wraps custom ``._fit_predict`` methods of blenders.

        Method wraps individual ``._fit_predict`` method of blenders.
        If input is single model - take it, else ``._fit_predict``
        Note - some pipelines may have more than 1 model.
        So corresponding prediction dataset have multiple prediction cols.

        Args:
            predictions: Sequence of datasets with predictions.
            pipes: Sequence of pipelines.
            class_mapping: Mapping for target classes.

        Returns:
            Single prediction dataset and sequence of pruned pipelines.

        """
        if len(pipes) == 1 and len(pipes[0].ml_algos) == 1:
            self._bypass = True
            return predictions[0], pipes

        return self._fit_predict(predictions, pipes, class_mapping)

    def _fit_predict(
        self, predictions: Sequence[LAMLDataset], pipes: Sequence[MLPipeline], class_mapping: Dict
    ) -> Tuple[LAMLDataset, Sequence[MLPipeline]]:
        """Defines how to fit, predict and prune - Abstract.

        Args:
            predictions: Sequence of datasets with predictions.
            pipes: Sequence of pipelines.
            class_mapping: Mapping for target classes.

        Returns:  # noqa: DAR202
            Single prediction dataset and sequence of pruned ``MLPipelines``.

        """
        raise NotImplementedError

    def predict(self, predictions: Sequence[LAMLDataset]) -> LAMLDataset:
        """Wraps custom ``._fit_predict`` methods of blenders.

        Args:
            predictions: Sequence of predictions from pruned datasets.

        Returns:
            Dataset with predictions.

        """
        if self._bypass:
            return predictions[0]

        return self._predict(predictions)

    def _predict(self, predictions: Sequence[LAMLDataset]) -> LAMLDataset:
        """Blend predictions on new sample.

        Args:
            predictions: Sequence of predictions from pruned datasets.

        Returns:  # noqa: DAR201
            Dataset with predictions.

        """
        raise NotImplementedError

    def split_models(self, predictions: Sequence[LAMLDataset]) -> Tuple[Sequence[LAMLDataset], List[int], List[int]]:
        """Split predictions by single model prediction datasets.

        Args:
            predictions: Sequence of datasets with predictions.

        Returns:
            Split predictions, model indices, pipe indices.

        """
        splitted_preds = []
        model_idx = []
        pipe_idx = []

        for n, preds in enumerate(predictions):

            features = preds.features
            n_models = len(features) // self.outp_dim

            for k in range(n_models):
                curr_pred = preds[:, features[k * self.outp_dim : (k + 1) * self.outp_dim]]
                splitted_preds.append(curr_pred)
                model_idx.append(k)
                pipe_idx.append(n)

        return splitted_preds, model_idx, pipe_idx

    def _set_metadata(self, predictions: Sequence[LAMLDataset], pipes: Sequence[MLPipeline], class_mapping: Dict):

        pred0 = predictions[0]
        pipe0 = pipes[0]

        self._outp_dim = pred0.shape[1] // len(pipe0.ml_algos)
        self._outp_prob = pred0.task.name in ["binary", "multiclass"]
        self._score = predictions[0].task.get_dataset_metric()

        self._class_mapping = class_mapping

    def score(self, dataset: LAMLDataset) -> float:
        """Score metric for blender.

        Args:
            dataset: Blended predictions dataset.

        Returns:
            Metric value.

        """
        return self._score(dataset, True)


class BestModelSelector(Blender):
    """Select best single model from level.

    Drops pipes that are not used in calc best model.
    Works in general case (even on some custom things)
    and most efficient on inference.
    Perform worse than other on tables,
    specially if some of models was terminated by timer.

    """

    def _fit_predict(
        self, predictions: Sequence[LAMLDataset], pipes: Sequence[MLPipeline], class_mapping: Dict
    ) -> Tuple[LAMLDataset, Sequence[MLPipeline]]:
        """Simple fit - just take one best.

        Args:
            predictions: Sequence of datasets with predictions.
            pipes: Sequence of pipelines.

        Returns:
            Single prediction dataset and Sequence of pruned pipelines.

        """
        self._set_metadata(predictions, pipes)
        splitted_preds, model_idx, pipe_idx = self.split_models(predictions)

        best_pred = None
        best_pipe_idx = 0
        best_model_idx = 0
        best_score = -np.inf

        for pred, mod, pipe in zip(splitted_preds, model_idx, pipe_idx):

            score = self.score(pred)

            if score > best_score:
                best_pipe_idx = pipe
                best_model_idx = mod
                best_score = score
                best_pred = pred

        best_pipe = pipes[best_pipe_idx]
        best_pipe.ml_algos = [best_pipe.ml_algos[best_model_idx]]

        return best_pred, [best_pipe]

    def _predict(self, predictions: Sequence[LAMLDataset]) -> LAMLDataset:
        """Simple predict - pruned pipe is a single model.

        Args:
            predictions: Sequence of predictions from pruned dataset.

        Returns:
            Dataset with predictions.

        """
        return predictions[0]


class MeanBlender(Blender):
    """Simple average level predictions.

    Works only with TabularDatasets.
    Doesn't require target to fit.
    No pruning.

    """

    def _get_mean_pred(self, splitted_preds: Sequence[NumpyDataset]) -> NumpyDataset:
        outp = splitted_preds[0].empty()

        pred = np.nanmean([x.data for x in splitted_preds], axis=0)

        outp.set_data(
            pred,
            ["MeanBlend_{0}".format(x) for x in range(pred.shape[1])],
            NumericRole(np.float32, prob=self._outp_prob),
        )

        return outp

    def _fit_predict(
        self, predictions: Sequence[NumpyDataset], pipes: Sequence[MLPipeline], class_mapping: Dict
    ) -> Tuple[NumpyDataset, Sequence[MLPipeline]]:
        """Simple fit_predict - just average and no prune.

        Args:
            predictions: Sequence of predictions.
            pipes: Sequence of pipelines.
            class_mapping: Mapping for target classes.

        Returns:
            Single prediction dataset and Sequence of pruned pipelines.

        """
        self._set_metadata(predictions, pipes, class_mapping)
        splitted_preds, _, __ = cast(List[NumpyDataset], self.split_models(predictions))

        outp = self._get_mean_pred(splitted_preds)

        return outp, pipes

    def _predict(self, predictions: Sequence[LAMLDataset]) -> LAMLDataset:
        """Simple fit_predict - just average.

        Args:
            predictions: Dataset with predictions.

        Returns:
            Dataset with averaged predictions.

        """
        splitted_preds, _, __ = cast(List[NumpyDataset], self.split_models(predictions))
        outp = self._get_mean_pred(splitted_preds)

        return outp


class WeightedBlender(Blender):
    """Weighted Blender based on coord descent, optimize task metric directly.

    Weight sum eq. 1.
    Good blender for tabular data,
    even if some predictions are NaN (ex. timeout).
    Model with low weights will be pruned.

    Args:
        max_iters: Max number of coord desc loops.
        max_inner_iters: Max number of iters to solve
            inner scalar optimization task.
        max_nonzero_coef: Maximum model weight value to stay in ensemble.

    """

    def __init__(
        self,
        max_iters: int = 5,
        max_inner_iters: int = 7,
        max_nonzero_coef: float = 0.05,
    ):
        self.max_iters = max_iters
        self.max_inner_iters = max_inner_iters
        self.max_nonzero_coef = max_nonzero_coef
        self.wts = [1]

    def _get_weighted_pred(self, splitted_preds: Sequence[NumpyDataset], wts: Optional[np.ndarray]) -> NumpyDataset:
        length = len(splitted_preds)
        if wts is None:
            wts = np.ones(length, dtype=np.float32) / length

        weighted_pred = np.nansum([x.data * w for (x, w) in zip(splitted_preds, wts)], axis=0).astype(np.float32)

        not_nulls = np.sum(
            [np.logical_not(np.isnan(x.data).any(axis=1)) * w for (x, w) in zip(splitted_preds, wts)],
            axis=0,
        ).astype(np.float32)

        not_nulls = not_nulls[:, np.newaxis]

        weighted_pred /= not_nulls
        weighted_pred = np.where(not_nulls == 0, np.nan, weighted_pred)

        outp = splitted_preds[0].empty()
        outp.set_data(
            weighted_pred,
            self._class_mapping if self._class_mapping else list(range(weighted_pred.shape[1])),
            NumericRole(np.float32, prob=self._outp_prob),
        )

        return outp

    def _get_candidate(self, wts: np.ndarray, idx: int, value: float) -> np.ndarray:
        """Replaces the value at the idx position in the input array with a new input value, \
        modifies the values in all other positions to keep the total sum of the array equal to 1.

        Args:
            wts (np.ndarray): input array
            idx (int): index
            value (float): new value

        Returns:
            np.ndarray: new array with sum equal to 1.
        """
        candidate = wts.copy()
        sl = np.arange(wts.shape[0]) != idx
        s = candidate[sl].sum()
        candidate[sl] = candidate[sl] / s * (1 - value)
        candidate[idx] = value

        # this is the part for pipeline pruning
        order = candidate.argsort()
        for idx in order:
            if candidate[idx] < self.max_nonzero_coef:
                candidate[idx] = 0
                candidate /= candidate.sum()
            else:
                break

        return candidate

    def _get_scorer(self, splitted_preds: Sequence[NumpyDataset], idx: int, wts: np.ndarray) -> Callable:
        def scorer(x):
            candidate = self._get_candidate(wts, idx, x)

            pred = self._get_weighted_pred(splitted_preds, candidate)
            score = self.score(pred)

            return -score

        return scorer

    def _optimize(self, splitted_preds: Sequence[NumpyDataset], eps: float = 1e-7) -> np.ndarray:
        """Perform coordinate descent.

        Args:
            splitted_preds (Sequence[NumpyDataset]): predictions for weighting to maximize the metric.
            eps (float): epsilon.

        Returns:
            np.ndarray: best weights.
        """
        # the set of initial weights for blending
        candidate = np.ones(len(splitted_preds), dtype=np.float32) / len(splitted_preds)

        best_score = self.score(self._get_weighted_pred(splitted_preds, candidate))
        best_weights = candidate
        logger.info(f"Blending: optimization starts with equal weights. Score = \x1b[1m{best_score:.7f}\x1b[0m")

        for iteration_num in range(self.max_iters):
            flg_no_upd = True
            for weights_idx in range(len(splitted_preds)):
                if candidate[weights_idx] == 1:
                    continue

                opt_res = minimize_scalar(
                    self._get_scorer(splitted_preds, weights_idx, candidate),
                    method="Bounded",
                    bounds=(0, 1),
                    options={"disp": False, "maxiter": self.max_inner_iters},
                )
                new_weight = opt_res.x
                score = -opt_res.fun
                candidate = self._get_candidate(candidate, weights_idx, new_weight)

                if score - eps > best_score:
                    flg_no_upd = False
                    best_score = score
                    best_weights = candidate

            if flg_no_upd:
                logger.info("Blending: no improvements for score. Terminated.\n")
                break
            else:
                logger.info(
                    f"Blending: iteration \x1b[1m{iteration_num}\x1b[0m: score = \x1b[1m{best_score:.7f}\x1b[0m, weights = \x1b[1m{best_weights}\x1b[0m"
                )

        logger.info(
            f"Blending: best score = \x1b[1m{best_score:.7f}\x1b[0m, best weights = \x1b[1m{best_weights}\x1b[0m"
        )

        return best_weights

    @staticmethod
    def _prune_pipe(
        pipes: Sequence[MLPipeline], wts: np.ndarray, pipe_idx: np.ndarray
    ) -> Tuple[Sequence[MLPipeline], np.ndarray]:
        new_pipes = []

        for i in range(max(pipe_idx) + 1):
            pipe = pipes[i]
            weights = wts[np.array(pipe_idx) == i]

            pipe.ml_algos = [x for (x, w) in zip(pipe.ml_algos, weights) if w > 0]

            new_pipes.append(pipe)

        new_pipes = [x for x in new_pipes if len(x.ml_algos) > 0]
        wts = wts[wts > 0]
        return new_pipes, wts

    def _fit_predict(
        self, predictions: Sequence[NumpyDataset], pipes: Sequence[MLPipeline], class_mapping: Dict
    ) -> Tuple[NumpyDataset, Sequence[MLPipeline]]:
        """Perform coordinate descent.

        Args:
            predictions: Sequence of prediction datasets.
            pipes: Sequence of pipelines.
            class_mapping: Mapping for target classes.

        Returns:
            Single prediction dataset and Sequence of pruned pipelines.

        Returns:
            Dataset and MLPipeline.

        """
        self._set_metadata(predictions, pipes, class_mapping)
        splitted_preds, _, pipe_idx = cast(List[NumpyDataset], self.split_models(predictions))

        wts = self._optimize(splitted_preds)
        splitted_preds = [x for (x, w) in zip(splitted_preds, wts) if w > 0]
        pipes, self.wts = self._prune_pipe(pipes, wts, pipe_idx)

        outp = self._get_weighted_pred(splitted_preds, self.wts)

        return outp, pipes

    def _predict(self, predictions: Sequence[LAMLDataset]) -> LAMLDataset:
        """Simple - weighted average.

        Args:
            predictions: Sequence of predictions.

        Returns:
            Dataset with weighted predictions.

        """
        splitted_preds, _, __ = cast(List[NumpyDataset], self.split_models(predictions))
        outp = self._get_weighted_pred(splitted_preds, self.wts)

        return outp



"""Base classes for metric and loss functions."""

from functools import partial
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np

from ..common_metric import _valid_str_metric_names
from ..utils import infer_gib


def fw_rmsle(x, y):
    """Function wrapper for rmsle."""
    return np.log1p(x), y


class MetricFunc:
    """Wrapper for metric.

    Args:
        metric_func: Callable metric function.
        m: Multiplier for metric value.
        bw_func: Backward function.

    """

    def __init__(self, metric_func, m, bw_func):
        self.metric_func = metric_func
        self.m = m
        self.bw_func = bw_func

    def __call__(self, y_true, y_pred, sample_weight=None) -> float:
        """Calculate metric."""
        y_pred = self.bw_func(y_pred)

        try:
            val = self.metric_func(y_true, y_pred, sample_weight=sample_weight)
        except TypeError:
            val = self.metric_func(y_true, y_pred)

        return val * self.m


class Loss:
    """Loss function with target transformation."""

    @staticmethod
    def _fw_func(target: Any, weights: Any) -> Tuple[Any, Any]:
        """Forward transformation.

        Args:
            target: Ground truth target values.
            weights: Item weights.

        Returns:
            Tuple (target, weights) without transformation.

        """
        return target, weights

    @staticmethod
    def _bw_func(pred: Any) -> Any:
        """Backward transformation for predicted values.

        Args:
            pred: Predicted target values.

        Returns:
            Pred without transformation.

        """
        return pred

    @property
    def fw_func(self):
        """Forward transformation for target values and item weights.

        Returns:
            Callable transformation.

        """
        return self._fw_func

    @property
    def bw_func(self):
        """Backward transformation for predicted values.

        Returns:
            Callable transformation.

        """
        return self._bw_func

    def metric_wrapper(
        self,
        metric_func: Callable,
        greater_is_better: Optional[bool],
        metric_params: Optional[Dict] = None,
    ) -> Callable:
        """Customize metric.

        Args:
            metric_func: Callable metric.
            greater_is_better: Whether or not higher value is better.
            metric_params: Additional metric parameters.

        Returns:
            Callable metric.

        """
        if greater_is_better is None:
            greater_is_better = infer_gib(metric_func)

        m = 2 * float(greater_is_better) - 1

        if metric_params is not None:
            metric_func = partial(metric_func, **metric_params)

        return MetricFunc(metric_func, m, self._bw_func)

    def set_callback_metric(
        self,
        metric: Union[str, Callable],
        greater_is_better: Optional[bool] = None,
        metric_params: Optional[Dict] = None,
        task_name: Optional[Dict] = None,
    ):
        """Callback metric setter.

        Args:
            metric: Callback metric
            greater_is_better: Whether or not higher value is better.
            metric_params: Additional metric parameters.
            task_name: Name of task.

        Note:
            Value of ``task_name`` should be one of following options:

            -  `'binary'`
            - `'reg'`
            - `'multiclass'`

        """
        assert task_name in [
            "binary",
            "reg",
            "multi:reg",
            "multiclass",
            "multilabel",
        ], "Incorrect task name: {}".format(task_name)
        self.metric = metric

        if metric_params is None:
            metric_params = {}

        if type(metric) is str:
            metric_dict = _valid_str_metric_names[task_name]
            self.metric_func = self.metric_wrapper(metric_dict[metric], greater_is_better, metric_params)
            self.metric_name = metric
        else:
            self.metric_func = self.metric_wrapper(metric, greater_is_better, metric_params)
            self.metric_name = None



# TODO: проверить что NLP не падает, CV
"""Tabular presets."""

import logging
import os

from collections import Counter
from copy import copy
from copy import deepcopy
from typing import Iterable
from typing import Optional
from typing import Sequence
from typing import cast

import numpy as np
import pandas as pd
import torch
import torch.nn as nn

from joblib import Parallel
from joblib import delayed
from pandas import DataFrame
from tqdm import tqdm

from ...addons.utilization import TimeUtilization
from ...dataset.np_pd_dataset import NumpyDataset
from ...ml_algo.boost_cb import BoostCB
from ...ml_algo.boost_lgbm import BoostLGBM
from ...ml_algo.boost_xgb import BoostXGB
from ...ml_algo.dl_model import TorchModel
from ...ml_algo.linear_sklearn import LinearLBFGS
from ...ml_algo.random_forest import RandomForestSklearn
from ...ml_algo.tuning.optuna import OptunaTuner
from ...pipelines.features.lgb_pipeline import LGBAdvancedPipeline
from ...pipelines.features.lgb_pipeline import LGBSeqSimpleFeatures
from ...pipelines.features.lgb_pipeline import LGBSimpleFeatures
from ...pipelines.features.linear_pipeline import LinearFeatures
from ...pipelines.features.torch_pipeline import TorchSimpleFeatures
from ...pipelines.ml.nested_ml_pipe import NestedTabularMLPipeline
from ...pipelines.selection.base import ComposedSelector
from ...pipelines.selection.base import SelectionPipeline
from ...pipelines.selection.importance_based import ImportanceCutoffSelector
from ...pipelines.selection.importance_based import ModelBasedImportanceEstimator
from ...pipelines.selection.permutation_importance_based import (
    NpIterativeFeatureSelector,
)
from ...pipelines.selection.permutation_importance_based import (
    NpPermutationImportanceEstimator,
)
from ...reader.base import DictToPandasSeqReader
from ...reader.base import PandasToPandasReader
from ...reader.tabular_batch_generator import ReadableToDf
from ...reader.tabular_batch_generator import read_batch
from ...reader.tabular_batch_generator import read_data
from ...tasks import Task
from ..blend import MeanBlender
from ..blend import WeightedBlender
from .base import AutoMLPreset
from .base import upd_params
from .utils import calc_feats_permutation_imps
from .utils import change_datetime
from .utils import plot_pdp_with_distribution


_base_dir = os.path.dirname(__file__)
logger = logging.getLogger(__name__)


class TabularAutoML(AutoMLPreset):
    """Classic preset - work with tabular data.

    Supported data roles - numbers, dates, categories.
    Limitations:

        - No memory management
        - No text support

    GPU support in catboost/lightgbm (if installed for GPU) training.

    Commonly _params kwargs (ex. timing_params) set via
    config file (config_path argument).
    If you need to change just few params, it's possible
    to pass it as dict of dicts, like json.
    To get available params please look on default config template.
    Also you can find there param description.
    To generate config template call
    :meth:`TabularAutoML.get_config('config_path.yml')`.

    Args:
        task: Task to solve.
        timeout: Timeout in seconds.
        memory_limit: Memory limit that are passed to each automl.
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.
        debug: To catch running model exceptions or not.
        timing_params: Timing param dict. Optional.
        config_path: Path to config file.
        general_params: General param dict.
        reader_params: Reader param dict.
        read_csv_params: Params to pass ``pandas.read_csv``
            (case of train/predict from file).
        nested_cv_params: Param dict for nested cross-validation.
        tuning_params: Params of Optuna tuner.
        selection_params: Params of feature selection.
        lgb_params: Params of lightgbm model.
        cb_params: Params of catboost model.
        xgb_params: Params of xgboost model.
        rf_params: Params of Sklearn Random Forest model.
        linear_l2_params: Params of linear model.
        nn_params: Params of neural network model.
        gbm_pipeline_params: Params of feature generation
            for boosting models.
        linear_pipeline_params: Params of feature generation
            for linear models.
        nn_pipeline_params: Params of feature generation
            for neural network models.
    """

    _default_config_path = "tabular_config.yml"

    # set initial runtime rate guess for first level models
    _time_scores = {
        "lgb": 1,
        "lgb_tuned": 3,
        "xgb": 1,
        "xgb_tuned": 3,
        "linear_l2": 0.7,
        "cb": 2,
        "cb_tuned": 6,
        "rf": 5,
        "rf_tuned": 10,
        "nn": 10,
        "nn_tuned": 20,
    }

    def __init__(
        self,
        task: Task,
        timeout: int = 3600,
        memory_limit: int = 16,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = "all",
        debug: bool = False,
        timing_params: Optional[dict] = None,
        config_path: Optional[str] = None,
        general_params: Optional[dict] = None,
        reader_params: Optional[dict] = None,
        read_csv_params: Optional[dict] = None,
        nested_cv_params: Optional[dict] = None,
        tuning_params: Optional[dict] = None,
        selection_params: Optional[dict] = None,
        lgb_params: Optional[dict] = None,
        cb_params: Optional[dict] = None,
        xgb_params: Optional[dict] = None,
        rf_params: Optional[dict] = None,
        linear_l2_params: Optional[dict] = None,
        nn_params: Optional[dict] = None,
        gbm_pipeline_params: Optional[dict] = None,
        linear_pipeline_params: Optional[dict] = None,
        nn_pipeline_params: Optional[dict] = None,
        time_series_pipeline_params: Optional[dict] = None,
        is_time_series: bool = False,
    ):
        super().__init__(task, timeout, memory_limit, cpu_limit, gpu_ids, debug, timing_params, config_path)
        #
        self.is_time_series = is_time_series

        # upd manual params
        for name, param in zip(
            [
                "general_params",
                "reader_params",
                "read_csv_params",
                "nested_cv_params",
                "tuning_params",
                "lgb_params",
                "cb_params",
                "xgb_params",
                "rf_params",
                "linear_l2_params",
                "nn_params",
                "gbm_pipeline_params",
                "linear_pipeline_params",
                "nn_pipeline_params",
            ],
            [
                general_params,
                reader_params,
                read_csv_params,
                nested_cv_params,
                tuning_params,
                lgb_params,
                cb_params,
                xgb_params,
                rf_params,
                linear_l2_params,
                nn_params,
                gbm_pipeline_params,
                linear_pipeline_params,
                nn_pipeline_params,
            ],
        ):
            if param is None:
                param = {}
            self.__dict__[name] = upd_params(self.__dict__[name], param)

        # if not time-series mode --> update selection_params too
        if not self.is_time_series:
            for name, param in zip(["selection_params"], [selection_params]):
                if param is None:
                    param = {}
                self.__dict__[name] = upd_params(self.__dict__[name], param)

        # if time-series mode --> update time_series_pipeline_params
        if self.is_time_series:
            for name, param in zip(["time_series_pipeline_params"], [time_series_pipeline_params]):
                if param is None:
                    param = {}
                self.__dict__[name] = upd_params(self.__dict__[name], param)

    def infer_auto_params(self, train_data: DataFrame, multilevel_avail: bool = False):

        length = train_data.shape[0]

        # infer optuna tuning iteration based on dataframe len
        if self.tuning_params["max_tuning_iter"] == "auto":
            if length < 10000:
                self.tuning_params["max_tuning_iter"] = 100
            elif length < 30000:
                self.tuning_params["max_tuning_iter"] = 50
            elif length < 100000:
                self.tuning_params["max_tuning_iter"] = 10
            else:
                self.tuning_params["max_tuning_iter"] = 5

        if self.general_params["use_algos"] == "auto":
            # TODO: More rules and add cases
            self.general_params["use_algos"] = [["lgb", "lgb_tuned", "linear_l2", "cb", "cb_tuned"]]
            if self.task.name == "multi:reg" and self.is_time_series:
                self.general_params["use_algos"] = [["cb", "linear_l2", "rf"]]
            else:
                if self.task.name == "multiclass" and multilevel_avail:
                    self.general_params["use_algos"].append(["linear_l2", "lgb"])

                if (self.task.name == "multi:reg") or (self.task.name == "multilabel"):
                    self.general_params["use_algos"] = [["linear_l2", "cb", "rf", "rf_tuned", "cb_tuned"]]

        if not self.general_params["nested_cv"]:
            self.nested_cv_params["cv"] = 1

        # check gpu to use catboost
        gpu_cnt = torch.cuda.device_count()
        gpu_ids = self.gpu_ids
        if gpu_cnt > 0 and gpu_ids:
            if gpu_ids == "all":
                gpu_ids = ",".join(list(map(str, range(gpu_cnt))))

            self.nn_params["device"] = gpu_ids.split(",")
            self.cb_params["default_params"]["task_type"] = "GPU"
            self.cb_params["default_params"]["devices"] = gpu_ids.replace(",", ":")

            self.xgb_params["default_params"]["device"] = f"cuda:{gpu_ids.split(',')[-1]}"  # TODO: add multigpu for xgb
        else:
            self.nn_params["device"] = "cpu"

        # check all n_jobs params
        cpu_cnt = min(os.cpu_count(), self.cpu_limit)
        torch.set_num_threads(cpu_cnt)

        self.cb_params["default_params"]["thread_count"] = min(
            self.cb_params["default_params"]["thread_count"], cpu_cnt
        )
        self.lgb_params["default_params"]["num_threads"] = min(
            self.lgb_params["default_params"]["num_threads"], cpu_cnt
        )
        self.xgb_params["default_params"]["nthread"] = min(self.xgb_params["default_params"]["nthread"], cpu_cnt)
        self.reader_params["n_jobs"] = min(self.reader_params["n_jobs"], cpu_cnt)

    def get_feature_pipeline(self, model, **kwargs):
        """Get LGBSeqSimpleFeatures pipeline if task is the time series prediction.

        Args:
            model: one from ["gbm", "linear_l2",, "rf", "nn"].
            kwargs: Arbitrary keyword arguments.

        Returns:
            appropriate features pipeline.
        """
        if self.is_time_series and model in ["gbm", "linear_l2", "rf", "nn"]:
            return LGBSeqSimpleFeatures(fill_na=True, scaler=True, transformers_params=self.time_series_pipeline_params)
        else:
            if model == "nn":
                return TorchSimpleFeatures(**self.nn_pipeline_params)
            if model == "linear_l2":
                return LinearFeatures(output_categories=True, **self.linear_pipeline_params)
            if model == "gbm":
                return LGBAdvancedPipeline(**self.gbm_pipeline_params, **kwargs)
            if model == "rf":
                if "fill_na" in kwargs:
                    return LGBAdvancedPipeline(**self.gbm_pipeline_params, **kwargs)
                return LGBAdvancedPipeline(**self.gbm_pipeline_params, fill_na=True, **kwargs)

    def get_time_score(self, n_level: int, model_type: str, nested: Optional[bool] = None):

        if nested is None:
            nested = self.general_params["nested_cv"]

        score = self._time_scores[model_type]

        mult = 1
        if nested:
            if self.nested_cv_params["n_folds"] is not None:
                mult = self.nested_cv_params["n_folds"]
            else:
                mult = self.nested_cv_params["cv"]

        if n_level > 1:
            mult *= 0.8 if self.general_params["skip_conn"] else 0.1

        score = score * mult

        # lower score for catboost on gpu
        if model_type in ["cb", "cb_tuned"] and self.cb_params["default_params"]["task_type"] == "GPU":
            score *= 0.5
        return score

    def get_selector(self, n_level: Optional[int] = 1) -> SelectionPipeline:
        selection_params = self.selection_params
        # lgb_params
        lgb_params = deepcopy(self.lgb_params)
        lgb_params["default_params"] = {
            **lgb_params["default_params"],
            **{"feature_fraction": 1},
        }

        cb_params = deepcopy(self.cb_params)
        cb_params["default_params"] = {
            **cb_params["default_params"],
            **{"rsm": 1},
        }

        mode = selection_params["mode"]

        # create pre selection based on mode
        pre_selector = None
        if mode > 0:
            # if we need selector - define model
            # timer will be useful to estimate time for next gbm runs
            selection_feats = LGBSimpleFeatures()

            if (self.task.name == "multi:reg") or (self.task.name == "multilabel"):
                time_score = self.get_time_score(n_level, "cb", False)
                sel_timer_0 = self.timer.get_task_timer("cb", time_score)
                selection_gbm = BoostCB(timer=sel_timer_0, **cb_params)
                model_name = "cb"
            else:
                time_score = self.get_time_score(n_level, "lgb", False)
                sel_timer_0 = self.timer.get_task_timer("lgb", time_score)
                selection_gbm = BoostLGBM(timer=sel_timer_0, **lgb_params)
                model_name = "lgb"
            selection_gbm.set_prefix("Selector")
            time_score = self.get_time_score(n_level, model_name, False)

            sel_timer_0 = self.timer.get_task_timer(model_name, time_score)

            if selection_params["importance_type"] == "permutation":
                importance = NpPermutationImportanceEstimator()
            else:
                importance = ModelBasedImportanceEstimator()

            pre_selector = ImportanceCutoffSelector(
                selection_feats,
                selection_gbm,
                importance,
                cutoff=selection_params["cutoff"],
                fit_on_holdout=selection_params["fit_on_holdout"],
            )
            if mode == 2:
                time_score = self.get_time_score(n_level, model_name, False)

                sel_timer_1 = self.timer.get_task_timer(model_name, time_score)
                selection_feats = LGBSimpleFeatures()
                if (self.task.name == "multi:reg") or (self.task.name == "multilabel"):
                    selection_gbm = BoostCB(timer=sel_timer_1, **cb_params)
                else:
                    selection_gbm = BoostLGBM(timer=sel_timer_1, **lgb_params)
                selection_gbm.set_prefix("Selector")

                # TODO: Check about reusing permutation importance
                importance = NpPermutationImportanceEstimator()

                extra_selector = NpIterativeFeatureSelector(
                    selection_feats,
                    selection_gbm,
                    importance,
                    feature_group_size=selection_params["feature_group_size"],
                    max_features_cnt_in_result=selection_params["max_features_cnt_in_result"],
                )

                pre_selector = ComposedSelector([pre_selector, extra_selector])

        return pre_selector

    def get_nn(
        self, keys: Sequence[str], n_level: int = 1, pre_selector: Optional[SelectionPipeline] = None
    ) -> NestedTabularMLPipeline:
        ml_algos = []
        force_calc = []

        nn_feats = self.get_feature_pipeline(model="nn")
        general_nn_params = deepcopy(self.nn_params)
        if "0" in self.nn_params:
            for i in range(len(keys)):
                if str(i) in general_nn_params:
                    del general_nn_params[str(i)]

        for i, key in enumerate(keys):
            time_score = self.get_time_score(n_level, "nn")
            nn_timer = self.timer.get_task_timer("reg_nn", time_score)
            model_params = deepcopy(general_nn_params)

            model_params.update(self.nn_params.get(str(i), general_nn_params))
            model_name = key

            if isinstance(key, str):
                tuned = "_tuned" in key
                if key[:2] == "nn":
                    model_name = "mlp"
                _name = "TorchNN_" + model_name + "_" + str(i)
                model_name = model_name.replace("_tuned", "")
            else:
                tuned = model_params.get("tuned")
                _name = "TorchNN_" + str(i)

            model_params["model"] = model_name
            nn_model = TorchModel(
                timer=nn_timer,
                default_params=model_params,
                freeze_defaults=model_params["freeze_defaults"],
                optimization_search_space=model_params.get("optimization_search_space", None),
            )
            nn_model._name = _name

            if tuned:
                nn_model.set_prefix("Tuned")
                nn_tuner = OptunaTuner(
                    n_trials=model_params["tuning_params"]["max_tuning_iter"],
                    timeout=model_params["tuning_params"]["max_tuning_time"],
                    fit_on_holdout=model_params["tuning_params"]["fit_on_holdout"],
                )
                nn_model = (nn_model, nn_tuner)
            ml_algos.append(nn_model)
            force_calc.append(True if not len(ml_algos) - 1 else False)

        nn_pipe = NestedTabularMLPipeline(
            ml_algos, force_calc, pre_selection=None, features_pipeline=nn_feats, **self.nested_cv_params
        )

        return nn_pipe

    def get_linear(self, n_level: int = 1, pre_selector: Optional[SelectionPipeline] = None) -> NestedTabularMLPipeline:

        # linear model with l2
        time_score = self.get_time_score(n_level, "linear_l2")
        linear_l2_timer = self.timer.get_task_timer("reg_l2", time_score)
        linear_l2_model = LinearLBFGS(timer=linear_l2_timer, **self.linear_l2_params)
        linear_l2_feats = self.get_feature_pipeline(model="linear_l2")

        linear_l2_pipe = NestedTabularMLPipeline(
            [linear_l2_model],
            force_calc=True,
            pre_selection=pre_selector,
            features_pipeline=linear_l2_feats,
            **self.nested_cv_params,
        )
        return linear_l2_pipe

    def get_gbms(
        self,
        keys: Sequence[str],
        n_level: int = 1,
        pre_selector: Optional[SelectionPipeline] = None,
    ):

        gbm_feats = self.get_feature_pipeline(model="gbm", feats_imp=pre_selector)

        ml_algos = []
        force_calc = []
        for idx, key in enumerate(keys):
            if idx == 0:  # force calculations only for the first ml_algo
                force = True

            tuned = "_tuned" in key
            algo_key = key.split("_")[0]
            time_score = self.get_time_score(n_level, key)
            gbm_timer = self.timer.get_task_timer(algo_key, time_score)
            if algo_key == "lgb":
                gbm_model = BoostLGBM(timer=gbm_timer, **self.lgb_params)
            elif algo_key == "cb":
                gbm_model = BoostCB(timer=gbm_timer, **self.cb_params)
            elif algo_key == "xgb":
                gbm_model = BoostXGB(timer=gbm_timer, **self.xgb_params)
            else:
                raise ValueError("Wrong algo key")

            if tuned:
                gbm_model.set_prefix("Tuned")
                gbm_tuner = OptunaTuner(
                    n_trials=self.tuning_params["max_tuning_iter"],
                    timeout=self.tuning_params["max_tuning_time"],
                    fit_on_holdout=self.tuning_params["fit_on_holdout"],
                )
                gbm_model = (gbm_model, gbm_tuner)
            ml_algos.append(gbm_model)
            force_calc.append(force)
            force = False

        gbm_pipe = NestedTabularMLPipeline(
            ml_algos, force_calc, pre_selection=pre_selector, features_pipeline=gbm_feats, **self.nested_cv_params
        )

        return gbm_pipe

    def get_rfs(self, keys: Sequence[str], n_level: int = 1, pre_selector: Optional[SelectionPipeline] = None):

        rf_feats = self.get_feature_pipeline(model="rf", feats_imp=pre_selector, fill_na=True)
        ml_algos = []
        force_calc = []
        for key, force in zip(keys, [True, False]):
            tuned = "_tuned" in key
            algo_key = key.split("_")[0]
            time_score = self.get_time_score(n_level, key)
            rf_timer = self.timer.get_task_timer(algo_key, time_score)

            rf_model = RandomForestSklearn(timer=rf_timer, **self.rf_params)

            if tuned:
                rf_model.set_prefix("Tuned")
                rf_tuner = OptunaTuner(
                    n_trials=self.tuning_params["max_tuning_iter"],
                    timeout=self.tuning_params["max_tuning_time"],
                    fit_on_holdout=self.tuning_params["fit_on_holdout"],
                )
                rf_model = (rf_model, rf_tuner)
            ml_algos.append(rf_model)
            force_calc.append(force)

        rf_pipe = NestedTabularMLPipeline(
            ml_algos, force_calc, pre_selection=pre_selector, features_pipeline=rf_feats, **self.nested_cv_params
        )

        return rf_pipe

    def create_automl(self, **fit_args):
        """Create basic automl instance.

        Args:
            **fit_args: Contain all information needed for creating automl.

        """
        train_data = fit_args["train_data"]
        multilevel_avail = fit_args["valid_data"] is None and fit_args["cv_iter"] is None

        if self.is_time_series:
            self.infer_auto_params(train_data["seq"]["seq0"], multilevel_avail)
            reader = DictToPandasSeqReader(task=self.task, **self.reader_params)
            pre_selector = None
        else:
            self.infer_auto_params(train_data, multilevel_avail)
            reader = PandasToPandasReader(task=self.task, **self.reader_params)
            pre_selector = self.get_selector()
        levels = []

        for n, names in enumerate(self.general_params["use_algos"]):
            lvl = []
            # regs
            rf_models = [x for x in ["rf", "rf_tuned"] if x in names]

            if len(rf_models) > 0:
                selector = None
                if (
                    self.is_time_series
                    or "rf" in self.selection_params["select_algos"]
                    and (self.general_params["skip_conn"] or n == 0)
                ):
                    selector = pre_selector
                lvl.append(self.get_rfs(rf_models, n + 1, selector))

            if "linear_l2" in names:
                selector = None
                if (
                    self.is_time_series
                    or "linear_l2" in self.selection_params["select_algos"]
                    and (self.general_params["skip_conn"] or n == 0)
                ):
                    selector = pre_selector
                lvl.append(self.get_linear(n + 1, selector))

            gbm_models = [
                x
                for x in ["lgb", "lgb_tuned", "cb", "cb_tuned", "xgb", "xgb_tuned"]
                if x in names and x.split("_")[0] in self.task.losses
            ]

            if len(gbm_models) > 0:
                selector = None
                if (
                    self.is_time_series
                    or "gbm" in self.selection_params["select_algos"]
                    and (self.general_params["skip_conn"] or n == 0)
                ):
                    selector = pre_selector
                lvl.append(self.get_gbms(gbm_models, n + 1, selector))

            available_nn_models = [
                "nn",
                "mlp",
                "dense",
                "denselight",
                "resnet",
                "snn",
                "linear_layer",
                "_linear_layer",
                "node",
                "autoint",
                "tabnet",
                "fttransformer",
            ]
            available_nn_models = available_nn_models + [x + "_tuned" for x in available_nn_models]
            nn_models = [
                x for x in names if x in available_nn_models or (isinstance(x, type) and issubclass(x, nn.Module))
            ]

            if len(nn_models) > 0:
                selector = None
                lvl.append(self.get_nn(nn_models, n + 1, selector))

            if len(lvl) != 0:
                levels.append(lvl)

        # blend everything
        blender = WeightedBlender(max_nonzero_coef=self.general_params["weighted_blender_max_nonzero_coef"])

        # initialize
        self._initialize(
            reader,
            levels,
            skip_conn=self.general_params["skip_conn"],
            blender=blender,
            return_all_predictions=self.general_params["return_all_predictions"],
            timer=self.timer,
            debug=self.debug,
        )

    def _get_read_csv_params(self):
        try:
            cols_to_read = self.reader.used_features
            numeric_dtypes = {
                x: self.reader.roles[x].dtype for x in self.reader.roles if self.reader.roles[x].name == "Numeric"
            }
        except AttributeError:
            cols_to_read = []
            numeric_dtypes = {}
        # cols_to_read is empty if reader is not fitted
        if len(cols_to_read) == 0:
            cols_to_read = None

        read_csv_params = copy(self.read_csv_params)
        read_csv_params = {
            **read_csv_params,
            **{"usecols": cols_to_read, "dtype": numeric_dtypes},
        }

        return read_csv_params

    def fit_predict(
        self,
        train_data: ReadableToDf,
        roles: Optional[dict] = None,
        train_features: Optional[Sequence[str]] = None,
        cv_iter: Optional[Iterable] = None,
        valid_data: Optional[ReadableToDf] = None,
        valid_features: Optional[Sequence[str]] = None,
        log_file: str = None,
        verbose: int = 0,
        path_to_save: Optional[str] = None,
    ) -> NumpyDataset:
        """Fit and get prediction on validation dataset.

        Almost same as :meth:`lightautoml.automl.base.AutoML.fit_predict`.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`.
              For example, ``{'data': X...}``. In this case,
              roles are optional, but `train_features`
              and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Args:
            train_data: Dataset to train.
            roles: Roles dict.
            train_features: Optional features names, if can't be inferred from `train_data`.
            cv_iter: Custom cv-iterator. For example, :class:`~lightautoml.validation.np_iterators.TimeSeriesIterator`.
            valid_data: Optional validation dataset.
            valid_features: Optional validation dataset features if cannot be inferred from `valid_data`.
            verbose: Controls the verbosity: the higher, the more messages.
                <1  : messages are not displayed;
                >=1 : the computation process for layers is displayed;
                >=2 : the information about folds processing is also displayed;
                >=3 : the hyperparameters optimization process is also displayed;
                >=4 : the training process for every algorithm is displayed;
            log_file: Filename for writing logging messages. If log_file is specified,
                the messages will be saved in a the file. If the file exists, it will be overwritten.
            path_to_save: The path that joblib will use to save the model after fit stage is completed. Use *.joblib format.

        Returns:
            Dataset with predictions. Call ``.data`` to get predictions array.

        """
        # roles may be none in case of train data is set {'data': np.ndarray, 'target': np.ndarray ...}
        self.set_logfile(log_file)
        if roles is None:
            roles = {}
        read_csv_params = self._get_read_csv_params()
        if self.is_time_series:
            train_data = train_data["seq"]["seq0"]
        train, upd_roles = read_data(train_data, train_features, self.cpu_limit, read_csv_params)
        if upd_roles:
            roles = {**roles, **upd_roles}
        if valid_data is not None:
            data, _ = read_data(valid_data, valid_features, self.cpu_limit, self.read_csv_params)

        if self.is_time_series:
            train = {"seq": {"seq0": train}}

        oof_pred = super().fit_predict(
            train,
            roles=roles,
            cv_iter=cv_iter,
            valid_data=valid_data,
            verbose=verbose,
            path_to_save=path_to_save,
        )

        return cast(NumpyDataset, oof_pred)

    def predict(
        self,
        data: ReadableToDf,
        features_names: Optional[Sequence[str]] = None,
        batch_size: Optional[int] = None,
        n_jobs: Optional[int] = 1,
        return_all_predictions: Optional[bool] = None,
    ) -> NumpyDataset:
        """Get dataset with predictions.

        Almost same as :meth:`lightautoml.automl.base.AutoML.predict`
        on new dataset, with additional features.

        Additional features - working with different data formats.
        Supported now:

            - Path to ``.csv``, ``.parquet``, ``.feather`` files.
            - :class:`~numpy.ndarray`, or dict of :class:`~numpy.ndarray`. For example,
              ``{'data': X...}``. In this case roles are optional,
              but `train_features` and `valid_features` required.
            - :class:`pandas.DataFrame`.

        Parallel inference - you can pass ``n_jobs`` to speedup
        prediction (requires more RAM).
        Batch_inference - you can pass ``batch_size``
        to decrease RAM usage (may be longer).

        Args:
            data: Dataset to perform inference.
            features_names: Optional features names,
                if cannot be inferred from `train_data`.
            batch_size: Batch size or ``None``.
            n_jobs: Number of jobs.
            return_all_predictions: if True,
                returns all model predictions from last level

        Returns:
            Dataset with predictions.

        """
        read_csv_params = self._get_read_csv_params()

        if batch_size is None and n_jobs == 1:
            if self.is_time_series:
                data = data["seq"]["seq0"]
            data, _ = read_data(data, features_names, self.cpu_limit, read_csv_params)
            if self.is_time_series:
                data = {"seq": {"seq0": data}}
            pred = super().predict(data, features_names, return_all_predictions)
            return cast(NumpyDataset, pred)

        data_generator = read_batch(
            data,
            features_names,
            n_jobs=n_jobs,
            batch_size=batch_size,
            read_csv_params=read_csv_params,
        )

        if n_jobs == 1:
            res = [self.predict(df, features_names, return_all_predictions) for df in data_generator]
        else:
            # TODO: Check here for pre_dispatch param
            with Parallel(n_jobs, pre_dispatch=len(data_generator) + 1) as p:
                res = p(delayed(self.predict)(df, features_names, return_all_predictions) for df in data_generator)

        res = NumpyDataset(
            np.concatenate([x.data for x in res], axis=0),
            features=res[0].features,
            roles=res[0].roles,
        )

        return res

    def get_feature_scores(
        self,
        calc_method: str = "fast",
        data: Optional[ReadableToDf] = None,
        features_names: Optional[Sequence[str]] = None,
        silent: bool = True,
    ):
        if calc_method == "fast":
            for level in self.levels:
                for pipe in level:
                    fi = pipe.pre_selection.get_features_score()
                    if fi is not None:
                        used_feats = set(self.collect_used_feats())
                        fi = fi.reset_index()
                        fi.columns = ["Feature", "Importance"]
                        return fi[fi["Feature"].map(lambda x: x in used_feats)]

            else:
                if not silent:
                    logger.info2("No feature importances to show. Please use another calculation method")
                return None

        if calc_method != "accurate":
            if not silent:
                logger.info2(
                    "Unknown calc_method. "
                    + "Currently supported methods for feature importances calculation are 'fast' and 'accurate'."
                )
            return None

        if data is None:
            if not silent:
                logger.info2("Data parameter is not setup for accurate calculation method. Aborting...")
            return None

        read_csv_params = self._get_read_csv_params()
        data, _ = read_data(data, features_names, self.cpu_limit, read_csv_params)
        used_feats = self.collect_used_feats()
        fi = calc_feats_permutation_imps(
            self,
            used_feats,
            data,
            self.reader.target,
            self.task.get_dataset_metric(),
            silent=silent,
        )
        return fi

    def get_individual_pdp(
        self,
        test_data: ReadableToDf,
        feature_name: str,
        n_bins: Optional[int] = 30,
        top_n_categories: Optional[int] = 10,
        datetime_level: Optional[str] = "year",
    ):
        assert feature_name in self.reader._roles
        assert datetime_level in ["year", "month", "dayofweek"]
        test_i = test_data.copy()
        # Numerical features
        if self.reader._roles[feature_name].name == "Numeric":
            counts, bin_edges = np.histogram(test_data[feature_name].dropna(), bins=n_bins)
            grid = (bin_edges[:-1] + bin_edges[1:]) / 2
            ys = []
            for i in tqdm(grid):
                test_i[feature_name] = i
                preds = self.predict(test_i).data
                ys.append(preds)
        # Categorical features
        if self.reader._roles[feature_name].name == "Category":
            feature_cnt = test_data[feature_name].value_counts()
            grid = list(feature_cnt.index.values[:top_n_categories])
            counts = list(feature_cnt.values[:top_n_categories])
            ys = []
            for i in tqdm(grid):
                test_i[feature_name] = i
                preds = self.predict(test_i).data
                ys.append(preds)
            if len(feature_cnt) > top_n_categories:
                freq_mapping = {feature_cnt.index[i]: i for i, _ in enumerate(feature_cnt)}
                # add "OTHER" class
                test_i = test_data.copy()
                # sample from other classes with the same distribution
                test_i[feature_name] = (
                    test_i[feature_name][np.array([freq_mapping[k] for k in test_i[feature_name]]) > top_n_categories]
                    .sample(n=test_data.shape[0], replace=True)
                    .values
                )
                preds = self.predict(test_i).data
                grid.append("<OTHER>")
                ys.append(preds)
                counts.append(feature_cnt.values[top_n_categories:].sum())
        # Datetime Features
        if self.reader._roles[feature_name].name == "Datetime":
            test_data_read = self.reader.read(test_data)
            feature_datetime = pd.arrays.DatetimeArray(test_data_read._data[feature_name])
            if datetime_level == "year":
                grid = np.unique([i.year for i in feature_datetime])
            elif datetime_level == "month":
                grid = np.arange(1, 13)
            else:
                grid = np.arange(7)
            ys = []
            for i in tqdm(grid):
                test_i[feature_name] = change_datetime(feature_datetime, datetime_level, i)
                preds = self.predict(test_i).data
                ys.append(preds)
            counts = Counter([getattr(i, datetime_level) for i in feature_datetime])
            counts = [counts[i] for i in grid]
        return grid, ys, counts

    def plot_pdp(
        self,
        test_data: ReadableToDf,
        feature_name: str,
        individual: Optional[bool] = False,
        n_bins: Optional[int] = 30,
        top_n_categories: Optional[int] = 10,
        top_n_classes: Optional[int] = 10,
        datetime_level: Optional[str] = "year",
    ):
        grid, ys, counts = self.get_individual_pdp(
            test_data=test_data,
            feature_name=feature_name,
            n_bins=n_bins,
            top_n_categories=top_n_categories,
            datetime_level=datetime_level,
        )
        plot_pdp_with_distribution(
            test_data,
            grid,
            ys,
            counts,
            self.reader,
            feature_name,
            individual,
            top_n_classes,
            datetime_level,
        )


class TabularUtilizedAutoML(TimeUtilization):
    """Template to make TimeUtilization from TabularAutoML.

    Simplifies using ``TimeUtilization`` module for ``TabularAutoMLPreset``.

    Args:
        task: Task to solve.
        timeout: Timeout in seconds.
        memory_limit: Memory limit that are passed to each automl.
        cpu_limit: CPU limit that that are passed to each automl.
        gpu_ids: GPU IDs that are passed to each automl.
        timing_params: Timing params level that are passed to each automl.
        configs_list: List of str path to configs files.
        drop_last: Usually last automl will be stopped with timeout.
            Flag that defines if we should drop it from ensemble.
        return_all_predictions: skip blending phase
        max_runs_per_config: Maximum number of multistart loops.
        random_state: Initial random seed that will be set
            in case of search in config.

    """

    def __init__(
        self,
        task: Task,
        timeout: int = 3600,
        memory_limit: int = 16,
        cpu_limit: int = 4,
        gpu_ids: Optional[str] = None,
        timing_params: Optional[dict] = None,
        configs_list: Optional[Sequence[str]] = None,
        drop_last: bool = True,
        return_all_predictions: bool = False,
        max_runs_per_config: int = 5,
        random_state: int = 42,
        outer_blender_max_nonzero_coef: float = 0.05,
        **kwargs,
    ):
        if configs_list is None:
            configs_list = [
                os.path.join(_base_dir, "tabular_configs", x)
                for x in [
                    "conf_0_sel_type_0.yml",
                    "conf_1_sel_type_1.yml",
                    "conf_2_select_mode_1_no_typ.yml",
                    "conf_3_sel_type_1_no_inter_lgbm.yml",
                    "conf_4_sel_type_0_no_int.yml",
                    "conf_5_sel_type_1_tuning_full.yml",
                    "conf_6_sel_type_1_tuning_full_no_int_lgbm.yml",
                ]
            ]
        inner_blend = MeanBlender()
        outer_blend = WeightedBlender(max_nonzero_coef=outer_blender_max_nonzero_coef)
        super().__init__(
            TabularAutoML,
            task,
            timeout,
            memory_limit,
            cpu_limit,
            gpu_ids,
            timing_params,
            configs_list,
            inner_blend,
            outer_blend,
            drop_last,
            return_all_predictions,
            max_runs_per_config,
            None,
            random_state,
            **kwargs,
        )

    def get_feature_scores(
        self,
        calc_method: str = "fast",
        data: Optional[ReadableToDf] = None,
        features_names: Optional[Sequence[str]] = None,
        silent: bool = True,
    ):
        if calc_method == "fast":
            feat_imps = []
            for pipe in self.outer_pipes:
                for model in pipe.ml_algos:
                    fi = model.models[0][0].get_feature_scores("fast")
                    if fi is not None:
                        feat_imps.append(fi)
            n_feat_imps = len(feat_imps)
            if n_feat_imps == 0:
                if not silent:
                    logger.info2(
                        "No feature importances to show. " + "Please use another calculation method or another preset."
                    )
                return None
            return (
                pd.concat(feat_imps).groupby("Feature")["Importance"].agg(sum).sort_values(ascending=False)
                / n_feat_imps
            ).reset_index()

        if calc_method != "accurate":
            if not silent:
                logger.info2(
                    "Unknown calc_method. "
                    + "Currently supported methods for feature importances calculation are 'fast' and 'accurate'."
                )
            return None

        if data is None:
            if not silent:
                logger.info2("Data parameter is not setup for accurate calculation method. Aborting...")
            return None

        automl = self.outer_pipes[0].ml_algos[0].models[0][0]
        read_csv_params = automl._get_read_csv_params()
        data, _ = read_data(data, features_names, self.cpu_limit, read_csv_params)

        used_feats = set()
        for pipe in self.outer_pipes:
            used_feats.update(pipe.ml_algos[0].models[0][0].collect_used_feats())

        fi = calc_feats_permutation_imps(
            automl,
            list(used_feats),
            data,
            automl.reader.target,
            automl.task.get_dataset_metric(),
            silent=silent,
        )
        return fi

    def create_model_str_desc(self, pref_tab_num: int = 0, split_line_len: int = 80) -> str:
        res = "Final prediction for new objects = \n"
        for it, (model, weight) in enumerate(zip(self.outer_pipes, self.outer_blend.wts)):
            config_path = model.ml_algos[0].models[0][0].config_path.split("/")[-1]
            res += "\t" * (pref_tab_num + 1) + "+ " * (it > 0)
            res += '{:.5f} * {} averaged models with config = "{}" and different CV random_states. Their structures: \n\n'.format(
                weight, len(model.ml_algos[0].models[0]), config_path
            )
            for it1, m in enumerate(model.ml_algos[0].models[0]):
                cur_model_desc = m.create_model_str_desc(pref_tab_num + 2, split_line_len)
                res += "\t" * (pref_tab_num + 1) + "    Model #{}.\n{}\n\n".format(it1, cur_model_desc)

        return res

    def get_individual_pdp(
        self,
        test_data: ReadableToDf,
        feature_name: str,
        n_bins: Optional[int] = 30,
        top_n_categories: Optional[int] = 10,
        datetime_level: Optional[str] = "year",
    ):
        reader = self.outer_pipes[0].ml_algos[0].models[0][0].reader
        assert feature_name in reader._roles
        assert datetime_level in ["year", "month", "dayofweek"]
        test_i = test_data.copy()
        # Numerical features
        if reader._roles[feature_name].name == "Numeric":
            counts, bin_edges = np.histogram(test_data[feature_name].dropna(), bins=n_bins)
            grid = (bin_edges[:-1] + bin_edges[1:]) / 2
            ys = []
            for i in tqdm(grid):
                test_i[feature_name] = i
                preds = self.predict(test_i).data
                ys.append(preds)
        # Categorical features
        if reader._roles[feature_name].name == "Category":
            feature_cnt = test_data[feature_name].value_counts()
            grid = list(feature_cnt.index.values[:top_n_categories])
            counts = list(feature_cnt.values[:top_n_categories])
            ys = []
            for i in tqdm(grid):
                test_i[feature_name] = i
                preds = self.predict(test_i).data
                ys.append(preds)
            if len(feature_cnt) > top_n_categories:
                freq_mapping = {feature_cnt.index[i]: i for i, _ in enumerate(feature_cnt)}
                # add "OTHER" class
                test_i = test_data.copy()
                # sample from other classes with the same distribution
                test_i[feature_name] = (
                    test_i[feature_name][np.array([freq_mapping[k] for k in test_i[feature_name]]) > top_n_categories]
                    .sample(n=test_data.shape[0], replace=True)
                    .values
                )
                preds = self.predict(test_i).data
                grid.append("<OTHER>")
                ys.append(preds)
                counts.append(feature_cnt.values[top_n_categories:].sum())
        # Datetime Features
        if reader._roles[feature_name].name == "Datetime":
            test_data_read = reader.read(test_data)
            feature_datetime = pd.arrays.DatetimeArray(test_data_read._data[feature_name])
            if datetime_level == "year":
                grid = np.unique([i.year for i in feature_datetime])
            elif datetime_level == "month":
                grid = np.arange(1, 13)
            else:
                grid = np.arange(7)
            ys = []
            for i in tqdm(grid):
                test_i[feature_name] = change_datetime(feature_datetime, datetime_level, i)
                preds = self.predict(test_i).data
                ys.append(preds)
            counts = Counter([getattr(i, datetime_level) for i in feature_datetime])
            counts = [counts[i] for i in grid]
        return grid, ys, counts

    def plot_pdp(
        self,
        test_data: ReadableToDf,
        feature_name: str,
        individual: Optional[bool] = False,
        n_bins: Optional[int] = 30,
        top_n_categories: Optional[int] = 10,
        top_n_classes: Optional[int] = 10,
        datetime_level: Optional[str] = "year",
    ):
        reader = self.outer_pipes[0].ml_algos[0].models[0][0].reader
        grid, ys, counts = self.get_individual_pdp(
            test_data=test_data,
            feature_name=feature_name,
            n_bins=n_bins,
            top_n_categories=top_n_categories,
            datetime_level=datetime_level,
        )
        plot_pdp_with_distribution(
            test_data,
            grid,
            ys,
            counts,
            reader,
            feature_name,
            individual,
            top_n_classes,
            datetime_level,
        )



"""Calculating inds for different TS datasets."""

import numpy as np
import pandas as pd

from ..dataset.roles import DatetimeRole


def sliding_window_view(a, window):
    """Generate window view."""
    shape = a.shape[:-1] + (a.shape[-1] - window + 1,) + (window,)
    strides = a.strides[:-1] + (a.strides[-1],) + a.strides[-1:]
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def rolling_window(a, window, step=1, from_last=True):
    """from_last == True - will cut first step-1 elements"""
    sliding_window = (
        sliding_window_view(a, window)
        if np.__version__ < "1.20"
        else np.lib.stride_tricks.sliding_window_view(a, window)
    )
    return sliding_window[(len(a) - window) % step if from_last else 0 :][::step]


class TopInd:
    def __init__(
        self, n_target=7, history=100, step=3, from_last=True, test_last=True, roles=None, scheme=None, **kwargs
    ):
        self.n_target = n_target
        self.history = history
        self.step = step
        self.from_last = from_last
        self.test_last = test_last
        self.roles = roles

    @staticmethod
    def _timedelta(x):
        delta = pd.to_datetime(x).diff().iloc[-1]
        if delta <= pd.Timedelta(days=1):
            return pd.to_datetime(x).diff().fillna(delta).values, delta

        if delta > pd.Timedelta(days=360):
            d = pd.to_datetime(x).dt.year.diff()
            delta = d.iloc[-1]
            return d.fillna(delta).values, delta
        elif delta > pd.Timedelta(days=27):
            d = pd.to_datetime(x).dt.month.diff() + 12 * pd.to_datetime(x).dt.year.diff()
            delta = d.iloc[-1]
            return d.fillna(delta).values, delta
        else:
            return pd.to_datetime(x).diff().fillna(delta).values, delta

    def read(self, data, plain_data=None):
        self.len_data = len(data)
        self.date_col = [col for col, role in self.roles.items() if isinstance(role, DatetimeRole)][0]
        self.time_delta = self._timedelta(data[self.date_col])[1]
        return self
        # TODO: add asserts

    def _create_test(self, data=None, plain_data=None):
        # for predicting future
        return rolling_window(
            np.arange(self.len_data if data is None else len(data)), self.history, self.step, self.from_last
        )[-1 if self.test_last else 0 :, :]

    def _create_data(self, data=None, plain_data=None):

        return rolling_window(
            np.arange(self.len_data if data is None else len(data))[: -self.n_target],
            self.history,
            self.step,
            self.from_last,
        )

    def _create_target(self, data=None, plain_data=None):
        return rolling_window(
            np.arange(self.len_data if data is None else len(data))[self.history :],
            self.n_target,
            self.step,
            self.from_last,
        )

    def _get_ids(self, data=None, plain_data=None, func=None, cond=None):
        date_col = pd.to_datetime(data[self.date_col])
        vals, time_delta = self._timedelta(data[self.date_col])
        ids = list(np.argwhere(vals != time_delta).flatten())
        prev = 0
        inds = []
        for split in ids + [len(date_col)]:
            segment = date_col.iloc[prev:split]
            if len(segment) > cond:
                ind = func(segment) + prev
                inds.append(ind)
            prev = split
        inds = np.vstack(inds)
        return inds

    def create_data(self, data=None, plain_data=None):
        return self._get_ids(data, plain_data, self._create_data, self.n_target + self.history)

    def create_test(self, data=None, plain_data=None):
        return self._get_ids(data, plain_data, self._create_test, self.history)

    def create_target(self, data=None, plain_data=None):
        return self._get_ids(data, plain_data, self._create_target, self.n_target + self.history)


class IDSInd:
    def __init__(self, scheme, **kwargs):
        self.scheme = scheme

    def read(self, data, plain_data=None):
        self.len_data = len(data)

        # TODO: add asserts

    def create_test(self, data, plain_data):
        # for predicting future
        return self.create_data(data, plain_data)

    def create_data(self, data, plain_data):
        ids = data.reset_index().groupby(self.scheme["from_id"])["index"].apply(self._to_list).to_dict()
        s = plain_data[self.scheme["to_id"]].map(ids)
        s.loc[s.isna()] = [[] for i in range(len(s.loc[s.isna()]))]
        return s.values

    def create_target(self, data=None, plain_data=None):
        return None

    @staticmethod
    def _to_list(x):
        return list(x)



"""Deep Learning transformers for calculating sentence embeddings."""

import gc

from copy import deepcopy
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence

import numpy as np
import torch
import torch.nn as nn

from sklearn.base import TransformerMixin
from torch.utils.data import DataLoader
from tqdm import tqdm


try:
    from transformers import AutoModel
except:
    import warnings

    warnings.warn("'transformers' - package isn't installed")

from ..ml_algo.torch_based.nn_models import SequenceAvgPooler
from ..ml_algo.torch_based.nn_models import SequenceClsPooler
from ..ml_algo.torch_based.nn_models import SequenceIndentityPooler
from ..ml_algo.torch_based.nn_models import SequenceMaxPooler
from ..ml_algo.torch_based.nn_models import SequenceSumPooler
from ..ml_algo.torch_based.nn_models import SequenceConcatPooler
from .dp_utils import CustomDataParallel
from .utils import _dtypes_mapping
from .utils import collate_dict
from .utils import parse_devices
from .utils import seed_everything
from .utils import single_text_hash


pooling_by_name = {
    "mean": SequenceAvgPooler,
    "sum": SequenceSumPooler,
    "max": SequenceMaxPooler,
    "concat": SequenceConcatPooler,
    "cls": SequenceClsPooler,
    "none": SequenceIndentityPooler,
}


class DLTransformer(TransformerMixin):
    """Deep Learning based sentence embeddings.

    Args:
        model: Torch model for aggregation word embeddings
            into sentence embedding.
        model_params: Dict with model parameters.
        dataset: Torch dataset.
        dataset_params: Dict with dataset params.
        loader_params: Dict with params for torch dataloader.
        device: String with torch device type or device ids. I.e: '0,2'.
        random_state: Determines random number generation.
        embedding_model: Torch word embedding model,
            if dataset do not return embeddings.
        embedding_model_params: Dict with embedding model params.
        multigpu: Use data parallel for multiple GPU.
        verbose: Show tqdm progress bar.

    """

    _model_params = {
        "embed_size": 300,
        "hidden_size": 256,
        "pooling": "mean",
        "num_layers": 1,
    }
    _loader_params = {"batch_size": 1024, "shuffle": False, "num_workers": 4}
    _dataset_params = {"embedding_model": None, "max_length": 200, "embed_size": 300}
    _embedding_model_params = {"model_name": "bert-base-cased"}

    def _infer_params(self):
        self.model_params = deepcopy(self._model_params)
        self.loader_params = deepcopy(self._loader_params)
        self.dataset_params = deepcopy(self._dataset_params)
        self.embedding_model_params = deepcopy(self._embedding_model_params)

    def __init__(
        self,
        model,
        model_params: Dict,
        dataset,
        dataset_params: Dict,
        loader_params: Dict,
        device: str = "cuda",
        random_state: int = 42,
        embedding_model: Optional[Any] = None,
        embedding_model_params: Dict[str, Dict] = None,
        multigpu: bool = False,
        verbose: bool = False,
    ):
        super(DLTransformer, self).__init__()
        self._infer_params()
        self.device, self.device_ids = parse_devices(device, multigpu)
        self.random_state = random_state
        self.verbose = verbose
        seed_everything(random_state)
        self.model_params.update(model_params)
        self.model = model(**self.model_params)

        self.embedding_model = None
        if embedding_model is not None:
            if embedding_model_params is not None:
                self.embedding_model_params.update(embedding_model_params)
            self.embedding_model = embedding_model(**self.embedding_model_params)

        self.dataset = dataset
        self.dataset_params.update(dataset_params)
        self.loader_params.update(loader_params)

    def get_name(self) -> str:
        """Module name.

        Returns:
            String with module name.

        """
        if self.embedding_model is None:
            name = self.model.get_name()
        else:
            name = self.model.get_name() + "_" + self.embedding_model.get_name()
        return name

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            Int with module output shape.

        """
        return self.model.get_out_shape()

    def fit(self, data: Any):  # noqa: D102
        return self

    @torch.no_grad()
    def transform(self, data: Sequence[str]) -> np.ndarray:
        """Embedded sentence."""
        dataset = self.dataset(data, **self.dataset_params)
        loader = DataLoader(dataset, collate_fn=collate_dict, **self.loader_params)

        result = []
        if self.verbose:
            loader = tqdm(loader)

        self.model = self.model.to(self.device)
        if self.device_ids is not None:
            self.model = CustomDataParallel(self.model, device_ids=self.device_ids)
        self.model.eval()

        if self.embedding_model is not None:
            self.embedding_model.to(self.device)
            if self.device_ids is not None:
                self.embedding_model = CustomDataParallel(self.embedding_model, device_ids=self.device_ids)
            self.embedding_model.eval()

        for sample in loader:
            data = {
                i: sample[i].long().to(self.device) if _dtypes_mapping[i] == "long" else sample[i].to(self.device)
                for i in sample.keys()
            }
            if self.embedding_model is not None:
                embed = self.embedding_model(data)
                if "attention_mask" in data:
                    length = torch.sum(data["attention_mask"], dim=1)
                else:
                    length = (torch.ones(len(embed)) * self.dataset_params["max_length"]).to(self.device).long()
                data = {"text": embed, "length": length}
            embed = self.model(data).detach().cpu().numpy()
            result.append(embed.astype(np.float32))

        result = np.vstack(result)
        self.model.to(torch.device("cpu"))
        if isinstance(self.model, CustomDataParallel):
            self.model = self.model.module

        if self.embedding_model is not None:
            self.embedding_model.to(torch.device("cpu"))

            if isinstance(self.embedding_model, CustomDataParallel):
                self.embedding_model = self.embedding_model.module

        del loader, dataset, data
        gc.collect()
        torch.cuda.empty_cache()

        return result


def position_encoding_init(n_pos: int, embed_size: int) -> torch.Tensor:
    """Compute positional embedding matrix.

    Args:
        n_pos: Len of sequence.
        embed_size: Size of output sentence embedding.

    Returns:
        Torch tensor with all positional embeddings.

    """
    position_enc = np.array(
        [
            [pos / np.power(10000, 2 * (j // 2) / embed_size) for j in range(embed_size)]
            if pos != 0
            else np.zeros(embed_size)
            for pos in range(n_pos)
        ]
    )
    position_enc[1:, 0::2] = np.sin(position_enc[1:, 0::2])
    position_enc[1:, 1::2] = np.cos(position_enc[1:, 1::2])
    return torch.from_numpy(position_enc).float()


class BOREP(nn.Module):
    """Class to compute Bag of Random Embedding Projections sentence embeddings from words embeddings.

    Bag of Random Embedding Projections sentence embeddings.

    Args:
        embed_size: Size of word embeddings.
        proj_size: Size of output sentence embedding.
        pooling: Pooling type.
        max_length: Maximum length of sentence.
        init: Type of weight initialization.
        pos_encoding: Add positional embedding.
        **kwargs: Ignored params.

    Note:
        There are several pooling types:

            - `'max'`: Maximum on seq_len dimension for non masked inputs.
            - `'mean'`: Mean on seq_len dimension for non masked inputs.
            - `'sum'`: Sum on seq_len dimension for non masked inputs.

        For init parameter there are several options:

            - `'orthogonal'`: Orthogonal init.
            - `'normal'`: Normal with std 0.1.
            - `'uniform'`: Uniform from -0.1 to 0.1.
            - `'kaiming'`: Uniform kaiming init.
            - `'xavier'`: Uniform xavier init.

    """

    name = "BOREP"
    _poolers = {"max", "mean", "sum"}

    def __init__(
        self,
        embed_size: int = 300,
        proj_size: int = 300,
        pooling: str = "mean",
        max_length: int = 200,
        init: str = "orthogonal",
        pos_encoding: bool = False,
        **kwargs: Any
    ):
        super(BOREP, self).__init__()
        self.embed_size = embed_size
        self.proj_size = proj_size
        self.pos_encoding = pos_encoding
        seed_everything(42)
        if self.pos_encoding:
            self.pos_code = position_encoding_init(max_length, self.embed_size).view(1, max_length, self.embed_size)

        self.pooling = pooling_by_name[pooling]()

        self.proj = nn.Linear(self.embed_size, self.proj_size, bias=False)

        if init == "orthogonal":
            nn.init.orthogonal_(self.proj.weight)
        elif init == "normal":
            nn.init.normal_(self.proj.weight, std=0.1)
        elif init == "uniform":
            nn.init.uniform_(self.proj.weight, a=-0.1, b=0.1)
        elif init == "kaiming":
            nn.init.kaiming_uniform_(self.proj.weight)
        elif init == "xavier":
            nn.init.xavier_uniform_(self.proj.weight)

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            Int with module output shape.

        """
        return self.proj_size

    def get_name(self) -> str:
        """Module name.

        Returns:
            String with module name.

        """
        return self.name

    @torch.no_grad()
    def forward(self, inp: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Forward-pass."""
        x = inp["text"]
        batch_size, batch_max_length = x.shape[0], x.shape[1]
        if self.pos_encoding:
            x = x + self.pos_code[:, :batch_max_length, :].to(x.device)
        x = x.contiguous().view(batch_size * batch_max_length, -1)
        x = self.proj(x)
        out = x.contiguous().view(batch_size, batch_max_length, -1)
        x_length = (torch.arange(out.shape[1])[None, :].to(out.device) < inp["length"][:, None])[:, :, None]
        out = self.pooling(out, x_length)

        return out


class RandomLSTM(nn.Module):
    """Class to compute Random LSTM sentence embeddings from words embeddings.

    Args:
        embed_size: Size of word embeddings.
        hidden_size: Size of hidden dimensions of LSTM.
        pooling: Pooling type.
        num_layers: Number of lstm layers.
        **kwargs: Ignored params.

    Note:
        There are several pooling types:

            - `'max'`: Maximum on seq_len dimension for non masked inputs.
            - `'mean'`: Mean on seq_len dimension for non masked inputs.
            - `'sum'`: Sum on seq_len dimension for non masked inputs.

    """

    name = "RandomLSTM"
    _poolers = ("max", "mean", "sum")

    def __init__(
        self, embed_size: int = 300, hidden_size: int = 256, pooling: str = "mean", num_layers: int = 1, **kwargs: Any
    ):
        super(RandomLSTM, self).__init__()
        if pooling not in self._poolers:
            raise ValueError("pooling - {} - not in the list of available types {}".format(pooling, self._poolers))
        seed_everything(42)
        self.hidden_size = hidden_size
        self.lstm = nn.LSTM(
            embed_size,
            hidden_size,
            num_layers=num_layers,
            bidirectional=True,
            batch_first=True,
        )

        self.pooling = pooling_by_name[pooling]()

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            Int with module output shape.

        """
        return self.hidden_size * 2

    def get_name(self) -> str:
        """Module name.

        Returns:
            String with module name.

        """
        return self.name

    @torch.no_grad()
    def forward(self, inp: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Forward-pass."""
        out, _ = self.lstm(inp["text"])
        x_length = (torch.arange(out.shape[1])[None, :].to(out.device) < inp["length"][:, None])[:, :, None]
        out = self.pooling(out, x_length)
        return out


class BertEmbedder(nn.Module):
    """Class to compute `HuggingFace <https://huggingface.co>`_ transformers words or sentence embeddings.

    Bert sentence or word embeddings.

    Args:
        model_name: Name of transformers model.
        pooling: Pooling type.
        **kwargs: Ignored params.

    Note:
        There are several pooling types:

            - `'cls'`: Use CLS token for sentence embedding
                from last hidden state.
            - `'max'`: Maximum on seq_len dimension
                for non masked inputs from last hidden state.
            - `'mean'`: Mean on seq_len dimension for non masked
                inputs from last hidden state.
            - `'sum'`: Sum on seq_len dimension for non masked inputs
                from last hidden state.
            - `'none'`: Don't use pooling (for RandomLSTM pooling strategy).

    """

    name = "BertEmb"
    _poolers = {"cls", "max", "mean", "sum", "none"}

    def __init__(self, model_name: str, pooling: str = "none", **kwargs: Any):
        super(BertEmbedder, self).__init__()
        if pooling not in self._poolers:
            raise ValueError("pooling - {} - not in the list of available types {}".format(pooling, self._poolers))

        self.pooling = pooling_by_name[pooling]()

        self.model_name = model_name
        self.transformer = AutoModel.from_pretrained(model_name)

    def forward(self, inp: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Forward-pass."""
        encoded_layers, _ = self.transformer(
            input_ids=inp["input_ids"],
            attention_mask=inp["attention_mask"],
            token_type_ids=inp.get("token_type_ids"),
            return_dict=False,
        )

        encoded_layers = self.pooling(encoded_layers, inp["attention_mask"].unsqueeze(-1).bool())

        return encoded_layers

    def freeze(self):
        """Freeze module parameters."""
        for param in self.transformer.parameters():
            param.requires_grad = False

    def get_name(self) -> str:
        """Module name.

        Returns:
            String with module name.

        """
        return self.name + single_text_hash(self.model_name)

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            Int with module output shape.

        """
        return self.transformer.config.hidden_size



"""Selectors for linear models."""

from typing import Optional
from typing import Union

import numpy as np

from scipy.sparse import linalg as sp_linalg

from ...validation.base import TrainValidIterator
from .base import SelectionPipeline


class HighCorrRemoval(SelectionPipeline):
    """Selector to remove highly correlated features.

    Del totally correlated feats to speedup L1 regression models.
    For sparse data cosine will be used.
    It's not exact, but ok for remove very high correlations.

    Args:
        corr_co: Similarity threshold.
        subsample: Number (int) of samples, or frac (float) from full dataset.
        random_state: Random seed for subsample.
        **kwargs: Additional parameters. Used for initialization of parent class.

    """

    def __init__(self, corr_co: float = 0.98, subsample: Union[int, float] = 100000, random_state: int = 42, **kwargs):
        super().__init__(**kwargs)
        self.corr_co = corr_co
        self.subsample = subsample
        self.random_state = random_state

    def perform_selection(self, train_valid: Optional[TrainValidIterator]):
        """Select features to save in dataset during selection.

        Method is used to perform selection based on features correlation.
        Should save ``_selected_features`` attribute in the end of working.

        Args:
            train_valid: Classic cv-iterator.

        """
        train = train_valid.train.data
        target = train_valid.train.target

        if train.shape[1] == 1:
            self._selected_features = train_valid.features
            return

        if self.subsample != 1 and self.subsample < train.shape[0]:
            if self.subsample < 1:
                subsample = int(train.shape[0] * self.subsample)
            else:
                subsample = int(self.subsample)

            idx = np.random.RandomState(self.random_state + 1).permutation(train.shape[0])[:subsample]
            train, target = train[idx], target[idx]

        # correlation or cosine
        if type(train) is np.ndarray:
            corr = np.corrcoef(train, rowvar=False)

        else:
            xtx = train.T * train
            norm = sp_linalg.norm(train, axis=0)
            corr = np.array(xtx / (norm[:, np.newaxis] * norm[np.newaxis, :]))
            del xtx

        sl = np.triu(np.abs(corr) > self.corr_co, k=1)
        grid_x, grid_y = np.meshgrid(np.arange(sl.shape[0]), np.arange(sl.shape[0]))

        removed = set()

        for x, y in zip(grid_x[sl], grid_y[sl]):
            if x not in removed:
                removed.add(y)

        const = np.arange(corr.shape[0])[np.isnan(np.diagonal(corr))]
        for i in const:
            removed.add(i)

        self._selected_features = [x for (n, x) in enumerate(train_valid.features) if n not in removed]



"""Neural Net modules for different data types."""

import logging

from typing import Any, List, Tuple, Type
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Union
from functools import reduce
import torch
import torch.nn as nn
from torch import Tensor
import operator
import numpy as np

try:
    from transformers import AutoModel
except:
    import warnings

    warnings.warn("'transformers' - package isn't installed")

from .dl_transformers import pooling_by_name


logger = logging.getLogger(__name__)


class TextBert(nn.Module):
    """Text data model.

    Class for working with text data based on HuggingFace transformers.

    Args:
        model_name: Transformers model name.
        pooling: Pooling type.

    Note:
        There are different pooling types:

            - cls: Use CLS token for sentence embedding
                from last hidden state.
            - max: Maximum on seq_len dimension for non masked
                inputs from last hidden state.
            - mean: Mean on seq_len dimension for non masked
                inputs from last hidden state.
            - sum: Sum on seq_len dimension for non masked
                inputs from last hidden state.
            - none: Without pooling for seq2seq models.

    """

    _poolers = {"cls", "max", "mean", "sum", "none"}

    def __init__(self, model_name: str = "bert-base-uncased", pooling: str = "cls"):
        super(TextBert, self).__init__()
        if pooling not in self._poolers:
            raise ValueError("pooling - {} - not in the list of available types {}".format(pooling, self._poolers))

        self.transformer = AutoModel.from_pretrained(model_name)
        self.n_out = self.transformer.config.hidden_size
        self.dropout = torch.nn.Dropout(0.2)
        self.activation = torch.nn.ReLU(inplace=True)
        self.pooling = pooling_by_name[pooling]()

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            int with module output shape.

        """
        return self.n_out

    def forward(self, inp: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Forward-pass."""
        # last hidden layer
        encoded_layers, _ = self.transformer(
            input_ids=inp["input_ids"],
            attention_mask=inp["attention_mask"],
            token_type_ids=inp.get("token_type_ids"),
            return_dict=False,
        )

        # pool the outputs into a vector
        encoded_layers = self.pooling(encoded_layers, inp["attention_mask"].unsqueeze(-1).bool())
        mean_last_hidden_state = self.activation(encoded_layers)
        mean_last_hidden_state = self.dropout(mean_last_hidden_state)
        return mean_last_hidden_state


class CatEmbedder(nn.Module):
    """Category data model.

    Args:
        cat_dims: Sequence with number of unique categories
            for category features.
        emb_dropout: Dropout probability.
        emb_ratio: Ratio for embedding size = (x + 1) // emb_ratio.
        max_emb_size: Max embedding size.

    """

    def __init__(
        self, cat_dims: Sequence[int], emb_dropout: bool = 0.1, emb_ratio: int = 3, max_emb_size: int = 50, **kwargs
    ):
        super(CatEmbedder, self).__init__()
        emb_dims = [(int(x), int(min(max_emb_size, max(1, (x + 1) // emb_ratio)))) for x in cat_dims]
        self.no_of_embs = sum([y for x, y in emb_dims])
        assert self.no_of_embs != 0, "The input is empty."
        # Embedding layers
        self.emb_layers = nn.ModuleList([nn.Embedding(x, y) for x, y in emb_dims])
        self.emb_dropout_layer = nn.Dropout(emb_dropout) if emb_dropout else nn.Identity()

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            Int with module output shape.

        """
        return self.no_of_embs

    def forward(self, inp: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Forward-pass."""
        output = torch.cat(
            [emb_layer(inp["cat"][:, i]) for i, emb_layer in enumerate(self.emb_layers)],
            dim=1,
        )
        output = self.emb_dropout_layer(output)
        return output


class ContEmbedder(nn.Module):
    """Numeric data model.

    Class for working with numeric data.

    Args:
        num_dims: Sequence with number of numeric features.
        input_bn: Use 1d batch norm for input data.

    """

    def __init__(self, num_dims: int, input_bn: bool = True, **kwargs):
        super(ContEmbedder, self).__init__()
        self.n_out = num_dims
        self.bn = nn.Identity()
        if input_bn:
            self.bn = nn.BatchNorm1d(num_dims)
        assert num_dims != 0, "The input is empty."

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            int with module output shape.

        """
        return self.n_out

    def forward(self, inp: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Forward-pass."""
        output = inp["cont"]
        output = self.bn(output)
        return output


class BasicCatEmbedding(nn.Module):
    """A basic embedding that creates an embedded vector for each field value from https://github.com/jrfiedler/xynn.

    The same as CatEmbedder, but without dropout, and it can be presented as a sequance.

    Args:
        embedding_size : int, optional
            size of each value's embedding vector; default is 10
        device : string or torch.device
        flatten_output: if flatten output or not.

    """

    def __init__(
        self,
        cat_dims: Sequence[int],
        embedding_size: int = 10,
        device: Union[str, torch.device] = "cuda:0",
        flatten_output: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.flatten_output = flatten_output
        self._device = device
        self.num_fields = 0
        self.output_size = 0
        self.embedding: Optional[nn.Embedding] = None
        self.embedding_size = embedding_size
        self._from_summary(cat_dims)
        self.cat_len = len(cat_dims)

    def _from_summary(self, cat_dims: Sequence[int]):
        num_values = 0

        self.emb_layers = nn.ModuleList([nn.Embedding(int(x), self.embedding_size) for x in cat_dims])
        self.num_fields = len(cat_dims)
        self.output_size = self.num_fields * self.embedding_size
        self.num_values = num_values
        for emb in self.emb_layers:
            nn.init.xavier_uniform_(emb.weight)

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            int with module output shape.

        """
        if self.flatten_output:
            return self.cat_len * self.embedding_size
        else:
            return self.cat_len

    def forward(self, X: Dict) -> Tensor:
        """Produce embedding for each value in input.

        Args:
            X : Dict

        Returns:
            torch.Tensor

        """
        X = X["cat"]
        x = torch.stack(
            [emb_layer(X[:, i]) for i, emb_layer in enumerate(self.emb_layers)],
            dim=1,
        )
        if self.flatten_output:
            return x.view(x.shape[0], -1)
        return x


class WeightedCatEmbedding(nn.Module):
    """DefaultEmbedding from https://github.com/jrfiedler/xynn.

    An embedding with a default value for each field. The default is returned for
    any field value not seen when the embedding was initialized (using `fit` or
    `from_summary`). For any value seen at initialization, a weighted average of
    that value's embedding and the default embedding is returned. The weights for
    the average are determined by the parameter `alpha`:

    weight = count / (count + alpha)
    final = embedding * weight + default * (1 - weight)

    Args:
        embedding_size : int, optional
            size of each value's embedding vector; default is 10
        alpha : int, optional
            controls the weighting of each embedding vector with the default;
            when `alpha`-many values are seen at initialization; the final
            vector is evenly weighted; the influence of the default is decreased
            with either higher counts or lower `alpha`; default is 20
        device : string or torch.device
        flatten_output: if flatten output or not.

    """

    def __init__(
        self,
        cat_dims: Sequence[int],
        cat_vc: Sequence[Dict],
        embedding_size: int = 10,
        alpha: int = 20,
        device: Union[str, torch.device] = "cuda:0",
        flatten_output: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.flatten_output = flatten_output
        self._device = device
        self.num_fields = 0
        self.output_size = 0
        self.alpha = alpha
        self.lookup: Dict[Tuple[int, Any], Tuple[int, int]] = {}
        self.lookup_default: Dict[int, Tuple[int, int]] = {}
        self.num_values = 0
        self.embedding: Optional[nn.Embedding] = None
        self.embedding_size = embedding_size
        self._from_summary(cat_vc, cat_dims)
        self.cat_len = len(cat_vc)
        self.cat_dims = cat_dims

    def _from_summary(self, unique_counts: List[Dict[Any, int]], cat_dims: Sequence[int]):
        self.emb_layers = nn.ModuleList([nn.Embedding(int(x), self.embedding_size) for x in cat_dims])
        self.def_layers = nn.ModuleList([nn.Embedding(1, self.embedding_size) for _ in cat_dims])
        weights_list = []
        for fieldnum, counts in enumerate(unique_counts):
            weights = []
            for i, vc in enumerate(sorted(counts.items())):
                value, count = vc
                if i == 0 and value != 0.0:
                    weights.append([0])
                weights.append([count / (count + self.alpha)])
            weights_list.append(weights)
        self.w_emb_layers = nn.ModuleList(
            [nn.Embedding.from_pretrained(torch.tensor(x, dtype=torch.float32)) for x in weights_list]
        )
        self.num_fields = len(unique_counts)
        self.output_size = self.num_fields * self.embedding_size
        for emb in self.emb_layers:
            nn.init.xavier_uniform_(emb.weight)

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            int with module output shape.

        """
        if self.flatten_output:
            return self.cat_len * self.embedding_size
        else:
            return self.cat_len

    def forward(self, X: Dict) -> Tensor:
        """Produce embedding for each value in input.

        Args:
            X : Dict

        Returns:
            torch.Tensor
        """
        X = X["cat"]
        emb_primary = torch.stack(
            [emb_layer(X[:, i]) for i, emb_layer in enumerate(self.emb_layers)],
            dim=1,
        )
        tsr_weights = torch.stack(
            [emb_layer(X[:, i]) for i, emb_layer in enumerate(self.w_emb_layers)],
            dim=1,
        )

        emb_default = torch.stack(
            [
                emb_layer(torch.tensor([0] * len(X[:, i]), device=self._device))
                for i, emb_layer in enumerate(self.def_layers)
            ],
            dim=1,
        )

        x = tsr_weights * emb_primary + (1 - tsr_weights) * emb_default
        if self.flatten_output:
            return x.view(x.shape[0], -1)
        return x


class LinearEmbedding(nn.Module):
    """An embedding for numeric fields from https://github.com/jrfiedler/xynn.

    There is one embedded vector for each field.
    The embedded vector for a value is that value times its field's vector.

    Args:
        embedding_size : int, optional
            size of each value's embedding vector; default is 10
        device : string or torch.device
        flatten_output: if flatten output or not.

    """

    def __init__(self, num_dims: int, embedding_size: int = 10, flatten_output: bool = False, **kwargs):
        super().__init__()
        self.flatten_output = flatten_output
        self.num_fields = num_dims
        self.output_size = 0
        self.embedding: Optional[nn.Embedding] = None
        self.embedding_size = embedding_size
        self._from_summary(self.num_fields)

    def _from_summary(self, num_fields: int):
        self.num_fields = num_fields
        self.output_size = num_fields * self.embedding_size
        self.embedding = nn.Embedding(num_fields, self.embedding_size)
        nn.init.xavier_uniform_(self.embedding.weight)

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            int with module output shape.

        """
        if self.flatten_output:
            return self.num_fields * self.embedding_size
        else:
            return self.num_fields

    def forward(self, X: Dict) -> Tensor:
        """Produce embedding for each value in input.

        Args:
            X : Dict

        Returns:
            torch.Tensor

        """
        X = X["cont"]
        x = self.embedding.weight * X.unsqueeze(dim=-1)
        if self.flatten_output:
            return x.view(x.shape[0], -1)
        return x


class DenseEmbedding(nn.Module):
    """An embedding for numeric fields, consisting of just a linear transformation with an activation from https://github.com/jrfiedler/xynn.

    Maps an input with shape n_rows * n_fields to an output with shape
    n_rows * 1 * embedding_size if one value passed for embedding_size or
    n_rows * embeddin_size[0] * embedding_size[1] if two values are passed

    Args:
        embedding_size : int, tuple of ints, or list of ints; optional
            size of each value's embedding vector; default is 10
        activation : subclass of torch.nn.Module, optional
            default is nn.LeakyReLU
        device : string or torch.device
        flatten_output: if flatten output or not.
    """

    def __init__(
        self,
        num_dims: int,
        embedding_size: Union[int, Tuple[int, ...], List[int]] = 10,
        activation: Type[nn.Module] = nn.LeakyReLU,
        flatten_output: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.flatten_output = flatten_output
        if isinstance(embedding_size, int):
            embedding_size = (1, embedding_size)
        elif len(embedding_size) == 1:
            embedding_size = (1, embedding_size[0])
        self.num_fields = num_dims
        self.output_size = 0
        self.embedding_w = None
        self.embedding_b = None
        self.dense_out_size = embedding_size
        self.embedding_size = embedding_size[-1]
        self.activation = activation()
        self._from_summary(self.num_fields)

    def _from_summary(self, num_fields: int):
        self.output_size = reduce(operator.mul, self.dense_out_size, 1)
        self.embedding_w = nn.Parameter(torch.zeros((num_fields, *self.dense_out_size)))
        self.embedding_b = nn.Parameter(torch.zeros(self.dense_out_size))
        nn.init.xavier_uniform_(self.embedding_w)

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            int with module output shape.

        """
        if self.flatten_output:
            return self.output_size
        else:
            return self.dense_out_size[0]

    def forward(self, X: Dict) -> Tensor:
        """Produce embedding for each value in input.

        Args:
            X : Dict

        Returns:
            torch.Tensor

        """
        X = X["cont"]
        embedded = self.embedding_w.T.matmul(X.T.float()).T + self.embedding_b
        embedded = self.activation(embedded.reshape((X.shape[0], -1)))
        x = embedded.reshape((X.shape[0], *self.dense_out_size))
        if self.flatten_output:
            return x.view(x.shape[0], -1)
        return x


class DenseEmbeddingFlat(DenseEmbedding):
    """Flatten version of DenseEmbedding."""

    def __init__(self, *args, **kwargs):
        super(DenseEmbeddingFlat, self).__init__(*args, **{**kwargs, **{"flatten_output": True}})


class LinearEmbeddingFlat(LinearEmbedding):
    """Flatten version of LinearEmbedding."""

    def __init__(self, *args, **kwargs):
        super(LinearEmbeddingFlat, self).__init__(*args, **{**kwargs, **{"flatten_output": True}})


class WeightedCatEmbeddingFlat(WeightedCatEmbedding):
    """Flatten version of WeightedCatEmbedding."""

    def __init__(self, *args, **kwargs):
        super(WeightedCatEmbeddingFlat, self).__init__(*args, **{**kwargs, **{"flatten_output": True}})


class BasicCatEmbeddingFlat(BasicCatEmbedding):
    """Flatten version of BasicCatEmbedding."""

    def __init__(self, *args, **kwargs):
        super(BasicCatEmbeddingFlat, self).__init__(*args, **{**kwargs, **{"flatten_output": True}})


class NLinearMemoryEfficient(nn.Module):
    """Linear multi-dim embedding from https://github.com/yandex-research/tabular-dl-num-embeddings/tree/c1d9eb63c0685b51d7e1bc081cdce6ffdb8886a8.

    Args:
        n : num of features.
        d_in: input size.
        d_out: output size.
    """

    def __init__(self, n: int, d_in: int, d_out: int) -> None:
        super().__init__()
        self.layers = nn.ModuleList([nn.Linear(d_in, d_out) for _ in range(n)])

    def forward(self, x):
        """Forward-pass."""
        return torch.stack([l(x[:, i]) for i, l in enumerate(self.layers)], 1)


class Periodic(nn.Module):
    """Periodic positional embedding for numeric features from https://github.com/yandex-research/tabular-dl-num-embeddings/tree/c1d9eb63c0685b51d7e1bc081cdce6ffdb8886a8.

    Args:
        n_features: num of numeric features
        emb_size: output size will be 2*emb_size
        sigma: weights will be initialized with N(0,sigma)
        flatten_output: if flatten output or not.
    """

    def __init__(
        self, n_features: int, emb_size: int = 64, sigma: float = 0.05, flatten_output: bool = False, **kwargs
    ) -> None:
        super().__init__()
        self.n_features = n_features
        self.emb_size = emb_size
        coefficients = torch.normal(0.0, sigma, (n_features, emb_size))
        self.coefficients = nn.Parameter(coefficients)
        self.flatten_output = flatten_output

    @staticmethod
    def _cos_sin(x: Tensor) -> Tensor:
        return torch.cat([torch.cos(x), torch.sin(x)], -1)

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            int with module output shape.

        """
        if self.flatten_output:
            return self.emb_size * 2 * self.n_features
        else:
            return self.n_features

    def forward(self, x: Tensor) -> Tensor:
        """Forward-pass."""
        x = self._cos_sin(2 * np.pi * self.coefficients[None] * x[..., None])
        if self.flatten_output:
            return x.view(x.shape[0], -1)
        return x


class PLREmbedding(nn.Module):
    """ReLU ◦ Linear ◦ Periodic embedding for numeric features from https://arxiv.org/pdf/2203.05556.pdf.

    Args:
        num_dims: int
        emb_size: int
        sigma: float
        flatten_output : bool
    """

    def __init__(
        self,
        num_dims: int,
        embedding_size: Union[int, Tuple[int, ...], List[int]] = 64,
        emb_size_periodic: int = 64,
        sigma_periodic: float = 0.05,
        flatten_output: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()
        self.num_dims = num_dims
        self.embedding_size = embedding_size
        self.layers: list[nn.Module] = []
        self.layers.append(Periodic(num_dims, emb_size_periodic, sigma_periodic))
        self.layers.append(NLinearMemoryEfficient(num_dims, 2 * emb_size_periodic, embedding_size))
        self.layers.append(nn.ReLU())
        self.layers = nn.Sequential(*self.layers)
        self.flatten_output = flatten_output

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            int with module output shape.

        """
        if self.flatten_output:
            return self.num_dims * self.embedding_size
        else:
            return self.num_dims

    def forward(self, X: Dict) -> Tensor:
        """Produce embedding for each value in input.

        Args:
            X : Dict

        Returns:
            torch.Tensor

        """
        X = X["cont"]
        x = self.layers(X)
        if self.flatten_output:
            return x.view(x.shape[0], -1)
        return x


class PLREmbeddingFlat(PLREmbedding):
    """Flatten version of BasicCatEmbedding."""

    def __init__(self, *args, **kwargs):
        super(PLREmbeddingFlat, self).__init__(*args, **{**kwargs, **{"flatten_output": True}})


class SoftEmbedding(torch.nn.Module):
    """Soft-one hot encoding embedding technique, from https://arxiv.org/pdf/1708.00065.pdf.

    In a nutshell, it represents a continuous feature as a weighted average of embeddings

    Args:
        num_embeddings: Number of embeddings to use (cardinality of the embedding table).
        embeddings_dim: The dimension of the vector space for projecting the scalar value.
        embeddings_init_std: The standard deviation factor for normal initialization of the
            embedding matrix weights.
        emb_initializer: Dict where keys are feature names and values are callable to initialize
            embedding tables
    """

    def __init__(self, num_dims, embedding_size=10, flatten_output: bool = False, **kwargs) -> None:
        super(SoftEmbedding, self).__init__()
        self.embedding_table = torch.nn.Embedding(num_dims, embedding_size)
        nn.init.xavier_uniform_(self.embedding_table.weight)

        self.projection_layer = torch.nn.Linear(1, num_dims, bias=True)
        self.softmax = torch.nn.Softmax(dim=-1)
        self.emb_size = embedding_size
        self.num_dims = num_dims
        self.flatten_output = flatten_output

    def get_out_shape(self) -> int:
        """Output shape.

        Returns:
            int with module output shape.

        """
        if self.flatten_output:
            return self.num_dims * self.emb_size
        else:
            return self.num_dims

    def forward(self, X: Dict) -> Tensor:
        """Produce embedding for each value in input.

        Args:
            X : Dict

        Returns:
            torch.Tensor

        """
        X = X["cont"]
        input_numeric = X.unsqueeze(-1)
        weights = self.softmax(self.projection_layer(input_numeric))
        x = (weights.unsqueeze(-1) * self.embedding_table.weight).sum(-2)
        if self.flatten_output:
            return x.view(x.shape[0], -1)
        return x


class SoftEmbeddingFlat(SoftEmbedding):
    """Flatten version of BasicCatEmbedding."""

    def __init__(self, *args, **kwargs):
        super(SoftEmbeddingFlat, self).__init__(*args, **{**kwargs, **{"flatten_output": True}})



"""GroupBy (categorical/numerical) features transformer."""

from ..dataset.roles import NumericRole
from ..pipelines.utils import get_columns_by_role
from .base import LAMLTransformer
from .utils import GroupByCatIsMode
from .utils import GroupByCatMode
from .utils import GroupByFactory
from .utils import GroupByNumDeltaMean
from .utils import GroupByNumDeltaMedian
from .utils import GroupByNumMax
from .utils import GroupByNumMin
from .utils import GroupByNumStd
from .utils import GroupByProcessor


class GroupByTransformer(LAMLTransformer):
    """Transformer, that calculates group_by features.

    Types of group_by features:
        - Group by categorical:
            - Numerical features:
                - Difference with group mode.
                - Difference with group median.
                - Group min.
                - Group max.
                - Group std.
            - Categorical features:
                - Group mode.
                - Is current value equal to group mode.

    Attributes:
        features list(str): generated features names.

    """

    _fit_checks = ()
    _transform_checks = ()
    _fname_prefix = "grb"

    @property
    def features(self):
        """Features list."""
        return self._features

    def __init__(self, num_groups=None, use_cat_groups=True, **kwargs):
        """Iniatialize.

        Args:
            num_groups: IDs of functions to use for numeric features.
            use_cat_groups: flag to show use for category features.
            **kwargs: additional params
        """
        super().__init__()

        self.num_groups = (
            num_groups
            if num_groups is not None
            else [
                GroupByNumDeltaMean.class_kind,
                GroupByNumDeltaMedian.class_kind,
                GroupByNumMin.class_kind,
                GroupByNumMax.class_kind,
                GroupByNumStd.class_kind,
            ]
        )
        self.use_cat_groups = use_cat_groups
        self.dicts = {}

    def fit(self, dataset):
        """Fit transformer and return it's instance.

        Args:
            dataset: Dataset to fit on.

        Returns:
            self.

        """
        # set transformer names and add checks
        for check_func in self._fit_checks:
            check_func(dataset)

        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()

        # set transformer features
        cat_cols = get_columns_by_role(dataset, "Category")
        num_cols = get_columns_by_role(dataset, "Numeric")
        feats = []
        for group_column in cat_cols:
            group_values = dataset.data[group_column].to_numpy()
            group_by_processor = GroupByProcessor(group_values)

            for feature_column in num_cols:
                for kind in self.num_groups:
                    feature = f"{self._fname_prefix}__{group_column}__{kind}__{feature_column}"
                    self.dicts[feature] = {
                        "group_column": group_column,
                        "feature_column": feature_column,
                        "groups": GroupByFactory.get_GroupBy(kind).fit(
                            data=dataset.data,
                            group_by_processor=group_by_processor,
                            feature_column=feature_column,
                        ),
                        "kind": kind,
                    }
                    feats.append(feature)

            if self.use_cat_groups:
                for feature_column in cat_cols:
                    if group_column != feature_column:
                        kind = GroupByCatMode.class_kind

                        # group results are the same for "cat_mode" and "cat_is_mode"
                        groups_1 = GroupByFactory.get_GroupBy(kind).fit(
                            data=dataset.data,
                            group_by_processor=group_by_processor,
                            feature_column=feature_column,
                        )

                        feature1 = f"{self._fname_prefix}__{group_column}__{kind}__{feature_column}"
                        self.dicts[feature1] = {
                            "group_column": group_column,
                            "feature_column": feature_column,
                            "groups": groups_1,
                            "kind": kind,
                        }

                        kind = GroupByCatIsMode.class_kind

                        # group results are the same for "cat_mode" and "cat_is_mode"
                        groups_2 = GroupByFactory.get_GroupBy(kind)
                        groups_2.set_dict(groups_1.get_dict())

                        feature2 = f"{self._fname_prefix}__{group_column}__{kind}__{feature_column}"
                        self.dicts[feature2] = {
                            "group_column": group_column,
                            "feature_column": feature_column,
                            "groups": groups_2,
                            "kind": kind,
                        }
                        feats.extend([feature1, feature2])

        self._features = feats

        return self

    def transform(self, dataset):
        """Calculate groups statistics.

        Args:
            dataset: Numpy or Pandas dataset with category and numeric columns.

        Returns:
            NumpyDataset of calculated group features (numeric).
        """
        # checks here
        super().transform(dataset)

        # convert to accepted dtype and get attributes
        dataset = dataset.to_pandas()

        # transform
        roles = NumericRole()
        outputs = []

        for feat, value in self.dicts.items():

            new_arr = value["groups"].transform(data=dataset.data, value=value)

            output = dataset.empty().to_numpy()
            output.set_data(new_arr, [feat], roles)
            outputs.append(output)

        # create resulted
        return dataset.empty().to_numpy().concat(outputs)
